# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import datetime
import functools
import json
from enum import Enum
from typing import Any
from typing import Callable
from typing import Dict
from typing import Union

import kolena
from kolena._api.v1.client_log import ClientLog as API
from kolena._api.v1.core import TestRun as CoreAPI
from kolena._api.v1.event import EventAPI
from kolena._utils import krequests


class DatadogLogLevels(str, Enum):
    ERROR = "error"


def upload_log(message: str, status: str) -> None:
    request = API.UploadLogRequest(
        client_version=kolena.__version__,
        timestamp=str(datetime.datetime.now()),
        message=message,
        status=status,
    )
    krequests.post(endpoint_path=API.Path.UPLOAD.value, json=dataclasses.asdict(request))


def report_crash(id: int, endpoint_path: str) -> None:
    request = CoreAPI.MarkCrashedRequest(test_run_id=id)
    # note no krequests.raise_for_status -- already in crashed state
    krequests.post(endpoint_path=endpoint_path, data=json.dumps(dataclasses.asdict(request)))


def set_profile() -> None:
    try:
        krequests.put(endpoint_path=EventAPI.Path.PROFILE)
    except Exception:
        """
        Attempting to set up event profile is best-effort. We don't want to have exceptions in that
        process be thrown to the customer--instead they should get their original stacktrace.
        """
        ...


def record_event(request: EventAPI.RecordEventRequest) -> None:
    try:
        krequests.post(endpoint_path=EventAPI.Path.EVENT, json=dataclasses.asdict(request))
    except Exception:
        """
        Attempting to record event is best-effort. We don't want to have exceptions in that
        process be thrown to the customer--instead they should get their original stacktrace.
        """
        ...


def with_event(event_name: str) -> Callable:
    """function decorator to track start and end of an event"""

    def event_decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            # track duration of the call, and if failed record the exception class name
            start_time = datetime.datetime.now()
            event_metadata: Dict[str, Union[str, int, float, bool, None]] = {}
            try:
                response = func(*args, **kwargs)
                return response
            except Exception as e:
                event_metadata["response_error"] = e.__class__.__name__
                raise e
            finally:
                event_metadata["duration"] = round((datetime.datetime.now() - start_time).total_seconds(), 3)
                record_event(
                    EventAPI.RecordEventRequest(event_name=event_name, additional_metadata=event_metadata),
                )

        return wrapper

    return event_decorator



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Special data types supported on the Kolena platform.

"""  # noqa: E501
from abc import ABCMeta
from datetime import datetime
from typing import Optional

import pytz

from kolena._utils.datatypes import DataCategory
from kolena._utils.datatypes import DataType
from kolena._utils.datatypes import TypedDataObject
from kolena._utils.pydantic_v1.dataclasses import dataclass
from kolena._utils.validators import ValidatorConfig


class _SpecialDataType(DataType):
    TIMESTAMP = "TIMESTAMP"

    @staticmethod
    def _data_category() -> DataCategory:
        return DataCategory.SPECIAL


@dataclass(frozen=True, config=ValidatorConfig)
class SpecialDataType(TypedDataObject[_SpecialDataType], metaclass=ABCMeta):
    """The base class for all special data types."""


@dataclass(frozen=True, config=ValidatorConfig)
class Timestamp(SpecialDataType):
    """
    !!! note "Experimental"
        This class is considered **experimental**

    Timestamp data type.
    """

    epoch_time: Optional[float] = None
    """The epoch time of the timestamp. If `value` and `format` are specified, the `epoch_time` will be calculated."""

    value: Optional[str] = None
    """
    The timestamp in a string representation. Note that GMT timezone is assumed unless the offset is specified in the
    string.
    """

    format: Optional[str] = None
    """
    The format of the `value` string following the
    [python format codes](https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes). If not
    provided, the `value` will be parsed using
    [python's `fromisoformat()`](https://docs.python.org/3/library/datetime.html#datetime.datetime.fromisoformat).
    """

    @staticmethod
    def _data_type() -> _SpecialDataType:
        return _SpecialDataType.TIMESTAMP

    def __post_init__(self) -> None:
        if self.value:
            if not self.format:
                time_obj = datetime.fromisoformat(self.value)
            else:
                time_obj = datetime.strptime(self.value, self.format)
            # assume GMT if timezone is not provided
            if not time_obj.tzinfo:
                time_obj = pytz.utc.localize(time_obj)
            object.__setattr__(self, "epoch_time", time_obj.timestamp())



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`kolena.io` provides helper functions for converting `pandas.DataFrame`s containing
to and from common serializable formats while adhering to the JSON specifications, and preserving
non-primitive data objects.
"""
import json
from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Union

import pandas as pd

from kolena._utils.dataframes.transformers import _try_parse
from kolena._utils.dataframes.transformers import df_apply
from kolena._utils.datatypes import _get_data_type
from kolena._utils.datatypes import DATA_TYPE_FIELD
from kolena._utils.datatypes import DataObject


class DataObjectJSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Dict:
        if isinstance(o, DataObject):
            return o._to_dict()
        return super().default(o)


def _deserialize_dataobject(x: Any) -> Any:
    if isinstance(x, list):
        return [_deserialize_dataobject(item) for item in x]

    if isinstance(x, dict):
        if data_type := x.pop(DATA_TYPE_FIELD, None):
            if typed_dataobject := _get_data_type(data_type):
                return typed_dataobject._from_dict(x)
        else:
            return {k: _deserialize_dataobject(v) for k, v in x.items()}

    return x


def _serialize_dataobject_str(x: Any) -> Any:
    if isinstance(x, (list, dict, DataObject)):
        return json.dumps(x, cls=DataObjectJSONEncoder)
    return x


def _deserialize_dataobject_str(x: Any) -> Any:
    y = _try_parse(x)
    return _deserialize_dataobject(y)


def dataframe_to_csv(df: pd.DataFrame, *args: Any, **kwargs: Any) -> Union[str, None]:
    """
    Helper function to export pandas DataFrame containing annotation or asset to CSV format.

    :param args: positional arguments to `pandas.DataFrame.to_csv`.
    :param kwargs: keyword arguments to `pandas.DataFrame.to_csv`.
    :return: None or str.
    """
    df_post = _dataframe_object_serde(df, _serialize_dataobject_str)
    return df_post.to_csv(*args, **kwargs)


def dataframe_from_csv(*args: Any, **kwargs: Any) -> pd.DataFrame:
    """
    Helper function to load pandas DataFrame exported to CSV with `dataframe_to_csv`.

    :param args: positional arguments to `pandas.DataFrame.read_csv`.
    :param kwargs: keyword arguments to `pandas.DataFrame.read_csv`.
    :return: DataFrame.
    """
    df = pd.read_csv(*args, **kwargs)
    df_post = _dataframe_object_serde(df, _deserialize_dataobject_str)

    return df_post


def dataframe_to_parquet(df: pd.DataFrame, *args: Any, **kwargs: Any) -> Union[bytes, None]:
    """
    Helper function to export pandas DataFrame containing annotation or asset to Parquet format.

    :param args: positional arguments to `pandas.DataFrame.to_parquet`.
    :param kwargs: keyword arguments to `pandas.DataFrame.to_parquet`.
    :return: None or str.
    """
    df_post = _dataframe_object_serde(df, partial(json.dumps, cls=DataObjectJSONEncoder))
    return df_post.to_parquet(*args, **kwargs)


def dataframe_from_parquet(*args: Any, **kwargs: Any) -> pd.DataFrame:
    """
    Helper function to load pandas DataFrame exported to Parquet with `dataframe_to_parquet`.

    :param args: positional arguments to `pandas.DataFrame.read_parquet`.
    :param kwargs: keyword arguments to `pandas.DataFrame.read_parquet`.
    :return: DataFrame.
    """
    df = pd.read_parquet(*args, **kwargs)
    df_post = _dataframe_object_serde(df, _deserialize_dataobject_str)

    return df_post


def dataframe_from_json(*args: Any, **kwargs: Any) -> pd.DataFrame:
    """
    Helper function to load pandas DataFrame containing annotation or asset from JSON file or string.

    :param args: positional arguments to `pandas.DataFrame.read_json`.
    :param kwargs: keyword arguments to `pandas.DataFrame.read_json`.
    :return: DataFrame.
    """
    df = pd.read_json(*args, **kwargs)
    df_post = _dataframe_object_serde(df, _deserialize_dataobject)

    return df_post


def _dataframe_object_serde(df: pd.DataFrame, serde_fn: Callable[[Any], Any]) -> pd.DataFrame:
    columns = list(df.columns)
    df_post = pd.DataFrame(columns=columns)
    for column in columns:
        if df.dtypes[column] == "object":
            df_post[column] = df_apply(df[column], serde_fn)
        else:
            df_post[column] = df[column]
    return df_post



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import root_validator
from pydantic import StrictFloat
from pydantic import StrictInt
from pydantic.dataclasses import dataclass
from pydantic.types import StrictBool
from pydantic.types import StrictStr
from typing_extensions import Literal


@dataclass(frozen=True)
class Range:
    min: float = -math.inf
    max: float = math.inf
    modulo: Optional[int] = None

    @classmethod
    @root_validator(skip_on_failure=True)
    def validate_min_max(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values["min"] > values["max"]:
            raise ValueError("Invalid min/max range.")

        return values


@dataclass
class FieldDescriptor:
    value_in: Optional[List[Union[StrictStr, StrictBool]]] = None
    number_range: Optional[Range] = None
    null_value: bool = False
    array_contains: Optional[List[Union[StrictStr, StrictBool, StrictInt, StrictFloat]]] = None
    value_in_distinct_count: Optional[int] = None

    def merge(self, other: "FieldDescriptor") -> "FieldDescriptor":
        return FieldDescriptor(
            value_in=self.value_in or other.value_in,
            number_range=self.number_range or other.number_range,
            null_value=self.null_value or other.null_value,
            array_contains=self.array_contains or other.array_contains,
            value_in_distinct_count=self.value_in_distinct_count or other.value_in_distinct_count,
        )

    def __post_init__(self) -> None:
        if self.value_in is not None and self.value_in_distinct_count is None:
            self.value_in_distinct_count = len(self.value_in)


@dataclass(frozen=True)
class FieldKey:
    key: str
    data: FieldDescriptor


@dataclass(frozen=True)
class GeneralFieldFilterOptions:
    regex_match: bool = False
    regex_flags: str = ""
    inverse: bool = False


@dataclass(frozen=True)
class GeneralFieldFilter:
    value_in: Optional[List[Union[StrictStr, StrictBool]]] = None
    number_range: Optional[Range] = None
    contains: Optional[str] = None
    null_value: Optional[Literal[True]] = None
    array_contains: Optional[List[Union[StrictStr, StrictBool, StrictInt, StrictFloat]]] = None
    options: GeneralFieldFilterOptions = GeneralFieldFilterOptions()

    @classmethod
    @root_validator(skip_on_failure=True)
    def validate_single_operation(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        # check exactly one filter operation is specified
        if sum(1 for value in values.values() if value is not None) != 1:
            raise ValueError("Must provide exactly one operation.")

        return values

    def __hash__(self) -> int:
        return hash((tuple(self.value_in or []), self.number_range, self.contains, self.null_value))



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Annotations are visualized in Kolena as overlays on top of [`TestSample`][kolena.workflow.TestSample] objects.

| Annotation | Valid [`TestSample`][kolena.workflow.TestSample] Types |
| --- | --- |
| [`BoundingBox`][kolena.workflow.annotation.BoundingBox] | [`Image`][kolena.workflow.Image], [`Video`][kolena.workflow.Video] |
| [`BoundingBox3D`][kolena.workflow.annotation.BoundingBox3D] | [`PointCloud`][kolena.workflow.PointCloud] |
| [`Polygon`][kolena.workflow.annotation.Polygon] | [`Image`][kolena.workflow.Image], [`Video`][kolena.workflow.Video] |
| [`Polyline`][kolena.workflow.annotation.Polyline] | [`Image`][kolena.workflow.Image], [`Video`][kolena.workflow.Video] |
| [`Keypoints`][kolena.workflow.annotation.Keypoints] | [`Image`][kolena.workflow.Image], [`Video`][kolena.workflow.Video] |
| [`SegmentationMask`][kolena.workflow.annotation.SegmentationMask] | [`Image`][kolena.workflow.Image], [`Video`][kolena.workflow.Video] |
| [`BitmapMask`][kolena.workflow.annotation.BitmapMask] | [`Image`][kolena.workflow.Image], [`Video`][kolena.workflow.Video] |
| [`Label`][kolena.workflow.annotation.Label] | [`Text`][kolena.workflow.Text], [`Document`][kolena.workflow.Document], [`Image`][kolena.workflow.Image], [`PointCloud`][kolena.workflow.PointCloud], [`Audio`][kolena.workflow.Audio], [`Video`][kolena.workflow.Video] |
| [`TimeSegment`][kolena.workflow.annotation.TimeSegment] | [`Audio`][kolena.workflow.Audio], [`Video`][kolena.workflow.Video] |

For example, when viewing images in the Studio, any annotations (such as lists of
[`BoundingBox`][kolena.workflow.annotation.BoundingBox] objects) present in the
[`TestSample`][kolena.workflow.TestSample], [`GroundTruth`][kolena.workflow.GroundTruth],
[`Inference`][kolena.workflow.Inference], or [`MetricsTestSample`][kolena.workflow.MetricsTestSample] objects are
rendered on top of the image.
"""  # noqa: E501
from kolena.annotation import Annotation
from kolena.annotation import BitmapMask
from kolena.annotation import BoundingBox
from kolena.annotation import BoundingBox3D
from kolena.annotation import ClassificationLabel
from kolena.annotation import Keypoints
from kolena.annotation import Label
from kolena.annotation import LabeledBoundingBox
from kolena.annotation import LabeledBoundingBox3D
from kolena.annotation import LabeledPolygon
from kolena.annotation import LabeledTimeSegment
from kolena.annotation import Polygon
from kolena.annotation import Polyline
from kolena.annotation import ScoredBoundingBox
from kolena.annotation import ScoredBoundingBox3D
from kolena.annotation import ScoredClassificationLabel
from kolena.annotation import ScoredLabel
from kolena.annotation import ScoredLabeledBoundingBox
from kolena.annotation import ScoredLabeledBoundingBox3D
from kolena.annotation import ScoredLabeledPolygon
from kolena.annotation import ScoredLabeledTimeSegment
from kolena.annotation import ScoredPolygon
from kolena.annotation import ScoredTimeSegment
from kolena.annotation import SegmentationMask
from kolena.annotation import TimeSegment

__all__ = [
    "Annotation",
    "BoundingBox",
    "LabeledBoundingBox",
    "ScoredBoundingBox",
    "ScoredLabeledBoundingBox",
    "Polygon",
    "LabeledPolygon",
    "ScoredPolygon",
    "ScoredLabeledPolygon",
    "Keypoints",
    "Polyline",
    "BoundingBox3D",
    "LabeledBoundingBox3D",
    "ScoredBoundingBox3D",
    "ScoredLabeledBoundingBox3D",
    "SegmentationMask",
    "BitmapMask",
    "Label",
    "ScoredLabel",
    "ClassificationLabel",
    "ScoredClassificationLabel",
    "TimeSegment",
    "LabeledTimeSegment",
    "ScoredTimeSegment",
    "ScoredLabeledTimeSegment",
]



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from typing import Literal
from typing import Optional

from pydantic import constr
from pydantic import model_validator
from pydantic.dataclasses import dataclass
from typing_extensions import Self

Source = Literal["datapoint", "result"]

VALID_NAME_PATTERN = r"^[A-Za-z0-9_]+$"


@dataclass(frozen=True)
class CreateRequest:
    dataset_id: int
    name: constr(min_length=1, strip_whitespace=True)
    formula: constr(min_length=1, strip_whitespace=True)
    # This field is intended for internal use and should be set by the system, not included in the request body
    source: Source = "datapoint"

    @model_validator(mode="after")
    def check_name(self) -> Self:
        name = self.name
        if not re.fullmatch(VALID_NAME_PATTERN, name):
            raise ValueError(f"Invalid name: {name}")
        return self


@dataclass(frozen=True)
class UpdateRequest:
    name: Optional[constr(min_length=1, strip_whitespace=True)] = None
    formula: Optional[constr(min_length=1, strip_whitespace=True)] = None
    # This field is intended for internal use and should be set by the system, not included in the request body
    source: Optional[Source] = None

    @model_validator(mode="after")
    def check_name(self) -> Self:
        name = self.name
        if name is None:
            return self

        if not re.fullmatch(VALID_NAME_PATTERN, name):
            raise ValueError(f"Invalid name: {name}")
        return self

    @model_validator(mode="after")
    def check_name_formula(self) -> Self:
        if self.name is None and self.formula is None:
            raise ValueError("Invalid request")
        return self


@dataclass(frozen=True)
class ListRequest:
    dataset_id: int


@dataclass(frozen=True)
class DerivedField:
    id: int
    dataset_id: int
    source: Source
    name: str
    formula: str

    def is_datapoint(self) -> bool:
        return self.source == "datapoint"

    def is_result(self) -> bool:
        return self.source == "result"



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
from dataclasses import field
from typing import Dict
from typing import List
from typing import Optional

from pydantic import conlist
from pydantic import model_validator
from pydantic.dataclasses import dataclass

from kolena._api.v2._api import GeneralFieldFilter
from kolena._api.v2._api import Range
from kolena._api.v2._derived_field import DerivedField
from kolena._api.v2._dsl import Dsl
from kolena._api.v2._metric import MetricGroup
from kolena.errors import IncorrectUsageError


@dataclass(frozen=True)
class ModelFilter:
    id: int
    eval_config_id: int
    result: Optional[Dict[str, GeneralFieldFilter]] = None
    is_null: Optional[bool] = None

    def __hash__(self) -> int:
        return hash(
            (
                self.id,
                self.eval_config_id,
                tuple(self.result or []),
                self.is_null,
            ),
        )

    @model_validator(mode="after")
    def validate_stratify_field_or_filter(self) -> "ModelFilter":
        if self.result and self.is_null:
            raise ValueError("Conflicting value filter and null filter.")
        return self


@dataclass(frozen=True)
class HumanEvaluationFilter:
    id: str
    human_evaluation: Optional[Dict[str, GeneralFieldFilter]] = None

    def __hash__(self) -> int:
        return hash(
            (
                self.id,
                tuple(self.human_evaluation or []),
            ),
        )


@dataclass(frozen=True)
class DifficultyScoreFilter:
    metric_groups: List[MetricGroup]

    def __hash__(self) -> int:
        return hash(
            (tuple(self.metric_groups),),
        )


class RelationshipOperator(str, enum.Enum):
    EQUAL = "equal"
    NOT_EQUAL = "not_equal"
    GREATER_THAN = "greater_than"
    SMALLER_THAN = "smaller_than"
    DIFFERENCE_RANGE = "difference_range"


class ComparisonObjectType(str, enum.Enum):
    RESULT = "result"
    DATAPOINT = "datapoint"

    def __str__(self) -> str:
        return self.value


@dataclass(frozen=True)
class ComparisonFilterDataSource:
    field: str
    type: ComparisonObjectType
    model_id: Optional[int] = None
    eval_config_id: Optional[int] = None


@dataclass(frozen=True)
class Relationship:
    operator: RelationshipOperator
    difference_range: Optional[Range] = None


@dataclass(frozen=True)
class CompareFilter:
    left: ComparisonFilterDataSource
    right: ComparisonFilterDataSource
    relationship: Relationship


@dataclass(frozen=True)
class Filters:
    dataset_ids: conlist(int, min_length=1)
    datapoint_ids: Optional[conlist(int, min_length=0)] = None
    datapoint: Dict[str, GeneralFieldFilter] = field(default_factory=dict)
    models: List[ModelFilter] = field(default_factory=list)
    compare_filters: List[CompareFilter] = field(default_factory=list)
    human_evaluations: conlist(HumanEvaluationFilter, max_length=1) = field(default_factory=list)
    difficulty_scores: Optional[DifficultyScoreFilter] = None
    dsl: Optional[Dsl] = None
    derived_fields: Optional[List[DerivedField]] = None

    def __hash__(self) -> int:
        return hash(
            (
                tuple(self.dataset_ids),
                tuple(self.datapoint),
                tuple(self.models),
                tuple(self.compare_filters),
                tuple(self.human_evaluations),
                self.difficulty_scores,
                self.dsl,
                tuple(self.derived_fields or []),
            ),
        )

    def __post_init__(self) -> None:
        all_evaluations = [(model.id, model.eval_config_id) for model in self.models]
        unique_evaluations = set(all_evaluations)
        if len(all_evaluations) != len(unique_evaluations):
            raise IncorrectUsageError("duplicate model and eval_config_id")
        if self.datapoint_ids:
            if len(set(self.datapoint_ids)) != len(self.datapoint_ids):
                raise IncorrectUsageError("duplicate datapoint ids")
        if self.difficulty_scores and not self.models:
            raise IncorrectUsageError("difficulty_scores cannot be used when models are not provided")



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atexit
import dataclasses
import functools
import inspect
import threading
import time
import uuid
from collections import defaultdict
from datetime import datetime
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import numpy as np
import pandas as pd

import kolena
from kolena._utils.datatypes import DataObject
from kolena.dataset import upload_results
from kolena.dataset.dataset import _load_dataset_metadata
from kolena.dataset.dataset import _upload_dataset
from kolena.errors import NotFoundError

THIRTY_SECONDS = 30
KOLENA_DEFAULT_ID = "_kolena_id"
KOLENA_TIMESTAMP_KEY = "_kolena_timestamp"
KOLENA_TIME_ELAPSED_KEY = "_kolena_time_elapsed"


def _serialize_item(value: Any):
    if isinstance(value, (bool, str, int, float)) or value is None:
        return value
    elif isinstance(value, np.generic):  # numpy scalars are common enough to be worth specific handling
        for base_type, numpy_type in [(bool, np.bool_), (int, np.integer), (float, np.inexact)]:
            if isinstance(value, numpy_type):  # cast if there is a match, otherwise fallthrough
                return base_type(value)
    elif isinstance(value, Dict):
        return {key: _serialize_item(subvalue) for key, subvalue in value.items()}
    elif isinstance(value, DataObject):
        return value._to_dict()
    elif dataclasses.is_dataclass(value):
        return _serialize_item(dataclasses.asdict(value))
    elif hasattr(value, "__dict__"):
        return _serialize_item(value.__dict__)
    elif isinstance(value, (List, Set, Tuple, pd.Series, np.ndarray)):
        return [_serialize_item(item) for item in value]
    else:
        return str(value)


class _Trace:
    def __init__(
        self,
        func: Callable,
        *,
        dataset_name: Optional[str] = None,
        model_name: Optional[str] = None,
        model_name_field: Optional[str] = None,
        sync_interval=THIRTY_SECONDS,
        id_fields: Optional[List[str]] = None,
        record_timestamp: bool = True,
    ):
        kolena.initialize(verbose=False)
        self.func = func
        self.signature = inspect.signature(func)
        if not dataset_name:
            dataset_name = func.__name__
        self.dataset_name = dataset_name
        if not model_name:
            model_name = f"{self.dataset_name}_model"
        self.model_name = model_name
        if model_name_field:
            if model_name_field not in self.signature.parameters:
                raise ValueError(f"Model Name Field {model_name_field} not found in function signature")
        self.model_name_field = model_name_field
        try:
            self.existing_dataset = _load_dataset_metadata(self.dataset_name)
            if not self.existing_dataset:
                raise NotFoundError("dataset metadata not found")
            if id_fields and sorted(id_fields) != sorted(self.existing_dataset.id_fields):
                raise ValueError(f"Id Fields {id_fields} do not match existing dataset id fields")
            self.id_fields = self.existing_dataset.id_fields
        except NotFoundError:
            self.existing_dataset = None
            if not id_fields:
                id_fields = [KOLENA_DEFAULT_ID]
            self.id_fields = id_fields
        for field in self.id_fields:
            if field not in self.signature.parameters and field != KOLENA_DEFAULT_ID:
                raise ValueError(f"Id Field {field} not found in function signature")
        self.datapoints = []
        self.results = defaultdict(list)
        self.last_update = time.time()
        self.task_ongoing = None
        self.sync_interval = sync_interval
        self.record_timestamp = record_timestamp
        functools.update_wrapper(self, func)
        atexit.register(self._clean_up)
        self.lock = threading.Lock()

    def _add_id_fields(self, datapoint: Dict, result: Dict) -> None:
        for field in self.id_fields:
            if field == KOLENA_DEFAULT_ID:
                call_id = uuid.uuid4().hex
                datapoint[KOLENA_DEFAULT_ID] = call_id
                result[KOLENA_DEFAULT_ID] = call_id
            else:
                field_value = datapoint.get(field)
                if field_value is None:
                    raise ValueError(f"Id Field {field} cannot be None in datapoint input")
                result[field] = field_value

    def __call__(self, *args, **kwargs):
        bounded_arguments = self.signature.bind(*args, **kwargs)
        bounded_arguments.apply_defaults()
        arguments = bounded_arguments.arguments

        start_time = datetime.now()
        output = self.func(**arguments)
        end_time = datetime.now()
        datapoint = _serialize_item(arguments)
        result = _serialize_item(output)
        if not isinstance(result, dict):
            result = {"result": result}
        if self.record_timestamp:
            datapoint[KOLENA_TIMESTAMP_KEY] = start_time.isoformat()
            result[KOLENA_TIMESTAMP_KEY] = end_time.isoformat()
            result[KOLENA_TIME_ELAPSED_KEY] = (end_time - start_time).total_seconds()
        self._add_id_fields(datapoint, result)
        model_name = self.model_name
        if self.model_name_field and arguments.get(self.model_name_field) is not None:
            model_name = arguments.get(self.model_name_field)
        with self.lock:
            self.datapoints.append(datapoint)
            self.results[model_name].append(result)
        if time.time() - self.last_update > self.sync_interval and (
            self.task_ongoing is None or not self.task_ongoing.is_alive()
        ):
            self.task_ongoing = threading.Thread(target=self._push_data)
            self.task_ongoing.start()
        return output

    def _push_data(self):
        try:
            with self.lock:
                datapoint_count = len(self.datapoints)
                results_counts = {model_name: len(results) for model_name, results in self.results.items()}
            dataset_df = pd.DataFrame(self.datapoints[:datapoint_count])
            unique_dataset_df = dataset_df.drop_duplicates(subset=self.id_fields)
            _upload_dataset(self.dataset_name, unique_dataset_df, id_fields=self.id_fields, append_only=True)
            with self.lock:
                self.datapoints = self.datapoints[dataset_df.shape[0] :]
            for model_name, results in self.results.items():
                result_df = pd.DataFrame(results[: results_counts[model_name]])
                upload_results(self.dataset_name, model_name, result_df.drop_duplicates(subset=self.id_fields))
                with self.lock:
                    self.results[model_name] = self.results[model_name][result_df.shape[0] :]
            self.last_update = time.time()
        except Exception as e:
            print(f"Failed to sync data: {e}")

    def _clean_up(self):
        if self.task_ongoing and self.task_ongoing.is_alive():
            self.task_ongoing.join()
        if self.datapoints or any(self.results.values()):
            self._push_data()


def kolena_trace(
    func: Optional[Callable] = None,
    *,
    dataset_name: Optional[str] = None,
    model_name: Optional[str] = None,
    model_name_field: Optional[str] = None,
    sync_interval: int = THIRTY_SECONDS,
    id_fields: Optional[List[str]] = None,
    record_timestamp: bool = True,
):
    """
    Use this decorator to trace the function with Kolena, the input and output of this function will be
    sent as datapoints and results respectively
    For example:
    ```python3
    @kolena_trace(dataset_name="test_trace", id_fields=["request_id"], record_timestamp=False)
    def predict(data, request_id):
        pass
    ```
    OR

    ```python3
    @kolena_trace
    def predict(data, request_id):
        pass
    ```
    :param func: The function to be traced, this is auto populated when used as a decorator
    :param dataset_name: The name of the dataset to be created, if not provided the function name will be used
    :param model_name: The name of the model to be created, if not provided the function name suffixed with _model
    will be used
    :param model_name_field: The field in the input that should be used as model name,
    if this would override the model name
    :param sync_interval: The interval at which the data should be synced to the server, default is 30 seconds
    :param id_fields: The fields in the input that should be used as id fields,
    if not provided a default id field will be used
    :param record_timestamp: If True, the timestamp of the input, output, and time elapsed will be recorded,
    default is True

    """
    if func:
        return _Trace(
            func,
            dataset_name=dataset_name,
            model_name=model_name,
            model_name_field=model_name_field,
            sync_interval=sync_interval,
            id_fields=id_fields,
            record_timestamp=record_timestamp,
        )
    else:

        def wrapper(func):
            return _Trace(
                func,
                dataset_name=dataset_name,
                model_name=model_name,
                model_name_field=model_name_field,
                sync_interval=sync_interval,
                id_fields=id_fields,
                record_timestamp=record_timestamp,
            )

        return wrapper



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from enum import Enum
from typing import List

from pydantic.v1 import conint

from kolena._api.v2._metric import MetricGroup
from kolena._api.v2._stratification import Stratification
from kolena._utils.pydantic_v1.dataclasses import dataclass


class Path(str, Enum):
    QUALITY_STANDARD = "quality-standard"
    RESULT = "quality-standard/result"
    COPY_FROM_DATASET = "quality-standard/copy-from-dataset"


@dataclass(frozen=True)
class CopyQualityStandardRequest:
    dataset_id: int
    source_dataset_id: int
    include_metric_groups: bool = True
    include_test_cases: bool = True


@dataclass(frozen=True)
class QualityStandard:
    name: str
    stratifications: List[Stratification]
    metric_groups: List[MetricGroup]

    @classmethod
    def metric_group_name_unique(cls, metric_groups: List[MetricGroup]) -> List[MetricGroup]:
        if len(metric_groups) > len({metric_group.name for metric_group in metric_groups}):
            raise ValueError("Metric group names must be unique.")
        return metric_groups


@dataclass(frozen=True)
class QualityStandardResponse:
    quality_standard_id: conint(gt=0)
    quality_standard: QualityStandard
    updated: datetime
    updated_by: str



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import asdict
from typing import List
from typing import Optional
from typing import Tuple

import pandas as pd

from kolena._api.v1.event import EventAPI
from kolena._api.v2.model import LoadByTagRequest
from kolena._api.v2.model import LoadByTagResponse
from kolena._api.v2.model import Path
from kolena._utils import krequests_v2 as krequests
from kolena._utils.instrumentation import with_event
from kolena._utils.serde import from_dict
from kolena.dataset.evaluation import download_results
from kolena.dataset.evaluation import EvalConfigResults
from kolena.errors import IncorrectUsageError
from kolena.errors import NotFoundError


def _get_models_by_tag(tag: str) -> List[str]:
    """
    Get model names by a given tag.

    :param tag: The tag on the model.
    :return: A list of model names associated with the provided tag.
    """
    response = krequests.put(
        Path.LOAD_BY_TAG,
        json=asdict(LoadByTagRequest(tag=tag)),
    )
    response.raise_for_status()

    load_by_tag_response = from_dict(LoadByTagResponse, response.json())
    return [model.name for model in load_by_tag_response.models]


@with_event(EventAPI.Event.FETCH_DATASET_MODEL_RESULT_BY_TAG)
def download_results_by_tag(
    dataset: str,
    model_tag: str,
    commit: Optional[str] = None,
    include_extracted_properties: bool = False,
) -> Tuple[pd.DataFrame, List[EvalConfigResults]]:
    """
    Download results given dataset name and model tag. Currently restricted to if the model tag is unique to the model.

    Concat dataset with results:

    ```python
    df_dp, results = download_results_by_tag("dataset name", "model tag")
    for eval_config, df_result in results:
        df_combined = pd.concat([df_dp, df_result], axis=1)
    ```

    :param dataset: The name of the dataset.
    :param model_tag: The tag associated with the model.
    :param commit: The commit hash for version control. Get the latest commit when this value is `None`.
    :param include_extracted_properties: If True, include kolena extracted properties from automated extractions
    in the datapoints and results as separate columns
    :return: Tuple of DataFrame of datapoints and list of [`EvalConfigResults`][kolena.dataset.EvalConfigResults].
    """
    model_names = _get_models_by_tag(model_tag)
    if not model_names:
        raise NotFoundError(f"no models with tag '{model_tag}' were found")
    elif len(model_names) > 1:
        raise IncorrectUsageError(f"multiple models with tag '{model_tag}' were found: {model_names}")
    else:
        return download_results(dataset, model_names[0], commit, include_extracted_properties)



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The ground truth associated with a [`TestSample`][kolena.workflow.TestSample]. Typically, a ground truth will represent
the expected output of a model when given a test sample and will be manually annotated by a human.

```python
from dataclasses import dataclass
from typing import List

from kolena.workflow import GroundTruth
from kolena.workflow.annotation import Polyline, SegmentationMask

@dataclass(frozen=True)
class AvGroundTruth(GroundTruth):
    road_area: SegmentationMask
    lane_boundaries: List[Polyline]
    visibility_score: int
```

A [`TestCase`][kolena.workflow.TestCase] holds a list of test samples (model inputs) paired with ground truths
(expected outputs).
"""
from typing import Type

from kolena._utils.datatypes import DataObject
from kolena._utils.pydantic_v1.dataclasses import dataclass
from kolena._utils.validators import ValidatorConfig
from kolena.workflow import Composite
from kolena.workflow import TestSample
from kolena.workflow._validators import get_data_object_field_types
from kolena.workflow._validators import safe_issubclass
from kolena.workflow._validators import validate_data_object_type
from kolena.workflow._validators import validate_field
from kolena.workflow.test_sample import _get_composite_fields


@dataclass(frozen=True, config=ValidatorConfig)
class GroundTruth(DataObject):
    """
    The ground truth against which a model is evaluated.

    A test case contains one or more [`TestSample`][kolena.workflow.TestSample] objects each paired with a ground truth
    object. During evaluation, these test samples, ground truths, and your model's inferences are provided to the
    [`Evaluator`][kolena.workflow.Evaluator] implementation.

    This object may contain any combination of scalars (e.g. `str`, `float`),
    [`Annotation`][kolena.workflow.annotation.Annotation] objects, or lists of these objects.

    For [`Composite`][kolena.workflow.Composite], each object can contain multiple basic test sample elements. To
    associate a set of attributes and/or annotations as the ground truth to a target test sample element, declare
    annotations by extending `DataObject` and use the same attribute name as used in the
    [`Composite`][kolena.workflow.Composite] test sample.

    Continue with the example given in [`Composite`][kolena.workflow.Composite], where the `FacePairSample` test sample
    type is defined using a pair of images under the `source` and `target` members, we can design a corresponding ground
    truth type with image-level annotations defined in the `FaceRegion` object:

    ```python
    from dataclasses import dataclass

    from kolena.workflow import DataObject, GroundTruth
    from kolena.workflow.annotation import BoundingBox, Keypoints

    @dataclass(frozen=True)
    class FaceRegion(DataObject):
        bounding_box: BoundingBox
        keypoints: Keypoints

    @dataclass(frozen=True)
    class FacePair(GroundTruth):
        source: FaceRegion
        target: FaceRegion
        is_same_person: bool
    ```

    This way, it is clear which bounding boxes and keypoints are associated to which image in the test sample.
    """


def _validate_ground_truth_type(
    test_sample_type: Type[TestSample],
    ground_truth_type: Type[GroundTruth],
) -> None:
    if not issubclass(ground_truth_type, GroundTruth):
        raise ValueError(f"Ground truth must subclass {GroundTruth.__name__}")

    is_composite = issubclass(test_sample_type, Composite)
    composite_fields = _get_composite_fields(test_sample_type) if is_composite else []

    for field_name, field_value in get_data_object_field_types(ground_truth_type).items():
        if field_name in composite_fields and safe_issubclass(field_value, DataObject):
            validate_data_object_type(field_value)
        else:
            validate_field(field_name, field_value)



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from dataclasses import astuple
from dataclasses import field
from enum import Enum
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

from pydantic import conint
from pydantic import conlist
from pydantic import constr
from pydantic import StrictInt
from pydantic import StrictStr
from pydantic.dataclasses import dataclass

from kolena._api.v2._filter import Filters
from kolena._api.v2._scope import Scopes
from kolena._api.v2._stats import SingleStatsResponse
from kolena.errors import IncorrectUsageError

MAX_BIN_COUNT = 200
MAX_STRATIFICATION_BUCKET_COUNT = 100
RESERVED_FILTER_FORMAT = "[]"


class Path(str, Enum):
    TESTING = "testing"


@dataclass(frozen=True)
class SimpleRange:
    min: Optional[float] = None
    max: Optional[float] = None

    def __post_init__(self) -> None:
        if self.min is not None and self.max is not None:
            if self.min > self.max:
                raise IncorrectUsageError("min value should be less than or equal to max")
        elif self.min is None and self.max is None:
            return

        if self.min is not None:
            if math.isnan(self.min):
                raise IncorrectUsageError("nan is not supported")
            if math.isinf(self.min):
                object.__setattr__(self, "min", None)

        if self.max is not None:
            if math.isnan(self.max):
                raise IncorrectUsageError("nan is not supported")
            if math.isinf(self.max):
                object.__setattr__(self, "max", None)

        if self.min is None and self.max is None:
            raise IncorrectUsageError(f"invalid values {self.min}, {self.max}")


Source = Literal["datapoint", "result"]


@dataclass(frozen=True)
class DatasetField:
    source: Source
    field: constr(min_length=1)

    def is_datapoint(self) -> bool:
        return self.source == "datapoint"

    def is_result(self) -> bool:
        return self.source == "result"


class StratificationType(str, Enum):
    CATEGORICAL = "categorical"
    EQUAL_HEIGHT = "equal-height"
    EQUAL_WIDTH = "equal-width"
    NUMERIC_INTERVAL = "numeric-interval"


@dataclass(frozen=True)
class StratificationInfo(DatasetField):
    value: Union[Optional[Union[StrictStr, StrictInt, bool, Tuple[Optional[float], Optional[float]]]], SimpleRange]
    index: int
    type: StratificationType

    def __post_init__(self) -> None:
        _validate_non_internal_list_field(self.field)

        if self.source == "result":
            fields = self.field.split(".", 2)
            if len(fields) >= 2:
                model_id = fields[0].strip('"')
                eval_config_id = fields[1].strip('"')
                object.__setattr__(self, "field", ".".join([model_id, eval_config_id, *fields[2:]]))

        if isinstance(self.value, tuple):
            if len(self.value) != 2:
                raise ValueError("invalid values", self.value)
            value = SimpleRange(min=self.value[0], max=self.value[1])
            object.__setattr__(self, "value", value)


@dataclass(frozen=True)
class TestCaseStats:
    sample_count: int
    metrics: SingleStatsResponse
    stratification: Optional[List[StratificationInfo]] = None


class BucketSplitType(str, Enum):
    EQUAL_WIDTH = "equal-width"
    EQUAL_HEIGHT = "equal-height"


@dataclass(frozen=True)
class TypedBucketSplit:
    count: conint(strict=True, gt=0, le=MAX_STRATIFICATION_BUCKET_COUNT) = 4
    type: BucketSplitType = BucketSplitType.EQUAL_HEIGHT


@dataclass(frozen=True)
class StratifyFieldSpec(DatasetField):
    values: Optional[List[Union[StrictStr, StrictInt, bool, None]]] = None
    buckets: Union[
        conlist(float, min_length=1, max_length=MAX_STRATIFICATION_BUCKET_COUNT),
        TypedBucketSplit,
        None,
    ] = None

    def __post_init__(self) -> None:
        _validate_non_internal_list_field(self.field)
        _validate_buckets(self.buckets)

    def is_categorical(self) -> bool:
        if self.values or self.buckets is None:
            return True

        return False

    def is_numeric(self) -> bool:
        return not self.is_categorical()


@dataclass(frozen=True)
class TestingRequest:
    filters: Filters
    stratify_fields: List[StratifyFieldSpec] = field(default_factory=list)
    scopes: Optional[Scopes] = None

    def __hash__(self) -> int:
        object_tuple = astuple(self, tuple_factory=tuple)
        serialized_str = str(object_tuple)
        return hash(serialized_str)


@dataclass(frozen=True)
class TestingResponse:
    dataset: TestCaseStats
    test_cases: List[TestCaseStats] = field(default_factory=list)


def _validate_non_internal_list_field(field: str) -> None:
    if RESERVED_FILTER_FORMAT in field:
        raise IncorrectUsageError("unsupported field")


def _validate_buckets(buckets: Union[List[float], TypedBucketSplit, None]) -> None:
    if isinstance(buckets, list):
        # allow [x, x] for single-value-single-bucket "stratification"
        if len(buckets) == 2 and math.isclose(buckets[0], buckets[1]):
            return
        if sorted(set(buckets)) != buckets:
            raise IncorrectUsageError("buckets must be strictly monotonically increasing")



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ._formula import accuracy
from ._formula import f1_score
from ._formula import fpr
from ._formula import precision
from ._formula import recall
from ._formula import specificity
from ._geometry import InferenceMatches
from ._geometry import iou
from ._geometry import match_inferences
from ._geometry import match_inferences_multiclass
from ._geometry import MulticlassInferenceMatches

__all__ = [
    "accuracy",
    "precision",
    "recall",
    "f1_score",
    "fpr",
    "specificity",
    "iou",
    "InferenceMatches",
    "match_inferences",
    "match_inferences_multiclass",
    "MulticlassInferenceMatches",
]



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandera as pa
from pandera.typing import Series

# Ensure check method is registered or else would get SchemaInitError
# noreorder
from kolena._utils.dataframes.validators import _validate_locator  # noqa: F401


class LocatorEmbeddingsDataFrameSchema(pa.DataFrameModel):
    key: Series[pa.typing.String] = pa.Field(coerce=True, _validate_locator=())
    """Unique key corresponding to model used for embeddings extraction. This is typically a locator."""

    locator: Series[pa.typing.String] = pa.Field(coerce=True, _validate_locator=())
    """External locator pointing to a sample in bucket."""

    embedding: Series[pa.typing.String] = pa.Field(coerce=True)
    """
    Embedding vector (base64-encoded string of `np.ndarray`) corresponding to a searchable representation of the sample.
    """


class DatasetEmbeddingsDataFrameSchema(pa.DataFrameModel):
    key: Series[pa.typing.String] = pa.Field(coerce=True)
    """
    Unique key corresponding  to the embedding vectors. This can be, for example, the name of the embedding model along
    with the column with which the embedding was extracted, such as "resnet50-image_locator".
    """

    datapoint_id_object: Series[pa.typing.String] = pa.Field(coerce=True)
    """
    String representation of the serialized datapoint id object from the dataset's id fields.
    """

    embedding: Series[pa.typing.String] = pa.Field(coerce=True)
    """
    Embedding vector (base64-encoded string of `np.ndarray`) corresponding to a searchable representation of the
        datapoint.
    """



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from collections import defaultdict
from dataclasses import asdict
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from kolena._api.v1.event import EventAPI
from kolena._api.v2._api import GeneralFieldFilter
from kolena._api.v2._filter import Filters
from kolena._api.v2._filter import ModelFilter
from kolena._api.v2._metric import Metric
from kolena._api.v2._testing import Path as TestingPath
from kolena._api.v2._testing import StratificationType
from kolena._api.v2._testing import StratifyFieldSpec
from kolena._api.v2._testing import TestingRequest
from kolena._api.v2._testing import TestingResponse
from kolena._api.v2.dataset import EntityData
from kolena._api.v2.model import ModelWithEvalConfig
from kolena._api.v2.quality_standard import CopyQualityStandardRequest
from kolena._api.v2.quality_standard import Path
from kolena._api.v2.quality_standard import QualityStandardResponse
from kolena._experimental.utils import get_delta_percentage
from kolena._experimental.utils import get_quantile
from kolena._experimental.utils import margin_of_error
from kolena._experimental.utils import ordinal
from kolena._utils import krequests_v2 as krequests
from kolena._utils import log
from kolena._utils.instrumentation import with_event
from kolena._utils.serde import from_dict
from kolena.dataset.dataset import _load_dataset_metadata
from kolena.dataset.evaluation import _get_eval_config_id
from kolena.dataset.evaluation import _get_model_id
from kolena.errors import IncorrectUsageError


PerformanceDelta = Literal["improved", "regressed", "similar", "unknown"]


def _format_quality_standard_result_df(quality_standard_result: dict) -> pd.DataFrame:
    df = pd.DataFrame.from_dict(quality_standard_result).rename_axis("stratification")

    df = df.explode("results_by_stratification")
    df = pd.concat(
        [df.drop(["results_by_stratification"], axis=1), df["results_by_stratification"].apply(pd.Series)],
        axis=1,
    )
    df.set_index(["test_case"], append=True, drop=True, inplace=True)

    df = df.explode("model_results")
    df["model"] = df["model_results"].apply(lambda x: x["model"]["name"])
    df["eval_config"] = df["model_results"].apply(lambda x: json.dumps(x["model"]["eval_config"]))
    df["metric_group"] = df["model_results"].apply(lambda x: list(x["results"].keys()))

    df = df.explode("metric_group")
    df["metric"] = df.apply(lambda x: list(x["model_results"]["results"][x["metric_group"]].keys()), axis=1)

    df = df.explode("metric")
    df["value"] = df.apply(lambda x: x["model_results"]["results"][x["metric_group"]][x["metric"]], axis=1)

    return df.pivot(columns=["model", "eval_config", "metric_group", "metric"], values="value")


def _download_quality_standard_result(
    dataset: str,
    models: List[str],
    metric_groups: Union[List[str], None] = None,
) -> pd.DataFrame:
    response = krequests.get(
        Path.RESULT,
        params=dict(dataset_name=dataset, models=models, metric_groups=metric_groups),
        api_version="v2",
    )
    krequests.raise_for_status(response)
    return _format_quality_standard_result_df(
        response.json(),
    )


def _download_test_case_result(
    dataset_id: int,
    model_and_eval_config_pairs: List[ModelWithEvalConfig],
    datapoint_filters: Optional[Dict[str, GeneralFieldFilter]],
    stratify_fields: List[StratifyFieldSpec],
) -> TestingResponse:
    filters = {}
    if datapoint_filters:
        for field, filter in datapoint_filters.items():
            if field.startswith("datapoint."):
                filters[field.replace("datapoint.", "")] = filter
            else:
                filters[field] = filter

    response = krequests.post(
        TestingPath.TESTING,
        json=asdict(
            TestingRequest(
                filters=Filters(
                    dataset_ids=[dataset_id],
                    datapoint=filters,
                    models=[
                        ModelFilter(
                            id=model.model_id,
                            eval_config_id=model.eval_config_id,
                        )
                        for model in model_and_eval_config_pairs
                    ],
                ),
                stratify_fields=stratify_fields,
            ),
        ),
        api_version="v2",
    )
    krequests.raise_for_status(response)
    return from_dict(TestingResponse, response.json())


def _download_quality_standard(
    dataset_id: int,
) -> QualityStandardResponse:
    response = krequests.get(
        Path.QUALITY_STANDARD,
        params=dict(dataset_id=dataset_id),
        api_version="v2",
    )
    krequests.raise_for_status(response)
    result = response.json()
    return from_dict(QualityStandardResponse, result)


def _calculate_moe_map(
    qs_result: pd.DataFrame,
    dataset_entity: EntityData,
    confidence_level: float,
    qs: QualityStandardResponse,
) -> Dict[Tuple[str, Any], float]:
    model_name_to_id_map = {}
    moe = {}
    for model, eval_config, _, _ in qs_result.columns:
        deserialized_eval_config = eval_config
        if eval_config == "null":
            deserialized_eval_config = None
        if eval_config and eval_config != "null":
            deserialized_eval_config = json.loads(eval_config)
        model_id = _get_model_id(model)
        eval_config_id = _get_eval_config_id(deserialized_eval_config)
        model_name_to_id_map[(model, eval_config)] = ModelWithEvalConfig(model_id, eval_config_id)

    for strat in qs.quality_standard.stratifications:
        test_cases = _download_test_case_result(
            dataset_entity.id,
            list(model_name_to_id_map.values()),
            strat.filters,
            strat.stratify_fields,
        )
        for i in range(len(test_cases.test_cases)):
            case = test_cases.test_cases[i]
            if not case.stratification or not case.stratification[0]:
                continue
            value = case.stratification[0].value
            field = case.stratification[0].field
            if field.startswith("_kolena.extracted"):
                field_keys = field.split(".")
                field = ".".join(field_keys[3:])
            if case.stratification[0].type in {StratificationType.EQUAL_HEIGHT, StratificationType.EQUAL_WIDTH}:
                value = f"{ordinal(i + 1)} {get_quantile(len(test_cases.test_cases))}"
            moe[(field, value)] = margin_of_error(case.sample_count, confidence_level)
        if ("Dataset", np.nan) not in moe:
            moe[("Dataset", np.nan)] = margin_of_error(test_cases.dataset.sample_count, confidence_level)
    return moe


def _get_performance_delta_metrics(
    qs: QualityStandardResponse,
    qs_result: pd.DataFrame,
    moe: Dict[Tuple[str, Any], float],
    reference_model: Optional[str] = None,
    reference_eval_config: Union[Dict[str, Any], str, None] = "first",
) -> pd.DataFrame:
    if not reference_eval_config:
        reference_eval_config = "null"
    for model, eval_config, _, _ in qs_result.columns:
        if not reference_model:
            reference_model = model
        if reference_model == model and reference_eval_config == "first":
            reference_eval_config = eval_config
    if isinstance(reference_eval_config, dict):
        reference_eval_config = json.dumps(reference_eval_config)
    performance_delta = pd.DataFrame("unknown", index=qs_result.index, columns=qs_result.columns)
    metric_dict: Dict[str, Dict[str, Metric]] = defaultdict(dict)
    for mg in qs.quality_standard.metric_groups:
        for metric in mg.metrics:
            metric_dict[mg.name][metric.label] = metric

    for model, eval_config, metric_group, metric_name in qs_result.columns:
        highlight = metric_dict[metric_group][metric_name].highlight
        if not highlight or highlight.higherIsBetter is None:
            continue

        for strat, test_case in qs_result.index:
            ref_val = qs_result.loc[
                (strat, test_case),
                (
                    reference_model,
                    reference_eval_config,
                    metric_group,
                    metric_name,
                ),
            ]
            target_val = qs_result.loc[(strat, test_case), (model, eval_config, metric_group, metric_name)]
            if (
                ref_val is None
                or target_val is None
                or np.isnan(ref_val)
                or np.isnan(target_val)
                or (strat, test_case) not in moe
            ):
                continue
            range_max = np.nanmax(
                qs_result.loc[strat][(reference_model, reference_eval_config, metric_group, metric_name)],
            )
            delta_percentage = get_delta_percentage(
                target_val,
                ref_val,
                range_max if range_max and not np.isnan(range_max) else max(target_val, ref_val),
            )
            if not highlight.higherIsBetter:
                delta_percentage = -delta_percentage
            if delta_percentage > moe[(strat, test_case)]:
                performance_delta.loc[
                    (strat, test_case),
                    (model, eval_config, metric_group, metric_name),
                ] = "improved"
            elif delta_percentage < -1 * moe[(strat, test_case)]:  # noqa: PAR001
                performance_delta.loc[
                    (strat, test_case),
                    (model, eval_config, metric_group, metric_name),
                ] = "regressed"
            else:
                performance_delta.loc[
                    (strat, test_case),
                    (model, eval_config, metric_group, metric_name),
                ] = "similar"
    return performance_delta


def _calculate_performance_delta(
    dataset: str,
    qs_result: pd.DataFrame,
    confidence_level: float,
    reference_model: Optional[str] = None,
    reference_eval_config: Union[Dict[str, Any], Literal["first"], None] = "first",
) -> pd.DataFrame:
    dataset_entity = _load_dataset_metadata(dataset)
    if not dataset_entity:
        raise IncorrectUsageError(f"The dataset with name '{dataset}' not found")
    qs = _download_quality_standard(dataset_entity.id)
    moe = _calculate_moe_map(qs_result, dataset_entity, confidence_level, qs)
    performance_delta = _get_performance_delta_metrics(qs, qs_result, moe, reference_model, reference_eval_config)

    qs_result = pd.concat(
        [qs_result, performance_delta],
        axis=1,
        keys=["metric_value", "performance_delta"],
        names=["type"],
    )
    qs_result = qs_result.reorder_levels(["model", "eval_config", "metric_group", "metric", "type"], axis=1)
    return qs_result


@with_event(event_name=EventAPI.Event.FETCH_QUALITY_STANDARD_RESULT)
def download_quality_standard_result(
    dataset: str,
    models: List[str],
    metric_groups: Union[List[str], None] = None,
    intersect_results: bool = True,
    confidence_level: Optional[float] = None,
    reference_model: Optional[str] = None,
    reference_eval_config: Optional[Union[Dict[str, Any], Literal["first"]]] = "first",
) -> pd.DataFrame:
    """
    Download quality standard result given a dataset and list of models.

    :param dataset: The name of the dataset.
    :param models: The names of the models.
    :param metric_groups: The names of the metric groups to include in the result.
    :param intersect_results: If True, only include datapoint that are common to all models in the metrics calculation.
    Note all metric groups are included when this value is `None`.
    :param confidence_level: The confidence score used to calculate the Margin of Error, representing the probability of
     capturing the true population parameter within the calculated MOE, we recommend setting between 0.9 - 0.99.
     If this is specified performance delta column will be shown, metric valued will be classified as
      improved, regressed, similar or unknown according to the MOE.
    :param reference_model: The name of the model to use as a reference for the Margin of Error calculation.
        This should be one of the models provided in `models`. If unspecified, the first model of `models` will be used.
    :param reference_eval_config: The evaluation configuration to use in conjunction with the reference model,
     if unspecified the first evaluation configuration of the reference model will be used
    :return: A Dataframe containing the quality standard result.
    """
    if reference_model and reference_model not in models:
        raise IncorrectUsageError(
            f"The specified reference model '{reference_model}' was not one of the provided models",
        )
    model_log = ", ".join([f"'{model}'" for model in models])
    log.info(f"downloading quality standard results for model(s) {model_log} on dataset '{dataset}'")
    if intersect_results:
        qs_result = _download_quality_standard_result(dataset, models, metric_groups)
    else:
        result_dfs = []
        for model in models:
            result_dfs.append(_download_quality_standard_result(dataset, [model], metric_groups))
        qs_result = pd.concat(result_dfs, axis=1)

    if confidence_level:
        qs_result = _calculate_performance_delta(
            dataset,
            qs_result,
            confidence_level,
            reference_model,
            reference_eval_config,
        )
    return qs_result


@with_event(event_name=EventAPI.Event.COPY_QUALITY_STANDARD_FROM_DATASET)
def copy_quality_standards_from_dataset(
    dataset: str,
    source_dataset: str,
    include_metric_groups: bool = True,
    include_test_cases: bool = True,
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """
    Create a quality standard on a dataset by copying from a source dataset. Note that this operation will overwrite the
    existing quality standards on the dataset if they exist.

    :param dataset: The name of the dataset.
    :param source_dataset: The name of the dataset from which the quality standards should be copied.
    :param include_metric_groups: Optional flag to indicate whether to copy the metric groups from the source dataset.
    :param include_test_cases: Optional flag to indicate whether to copy the test cases from the source dataset.
    :return: A tuple of the created metric groups and test cases.
    """
    if dataset == source_dataset:
        raise IncorrectUsageError("source dataset and target dataset are the same")

    if not include_test_cases and not include_metric_groups:
        raise IncorrectUsageError("should include at least one of metric group or test case")

    dataset_metadata = _load_dataset_metadata(dataset)
    if not dataset_metadata:
        raise IncorrectUsageError(f"The dataset with name '{dataset}' not found")
    source_dataset_metadata = _load_dataset_metadata(source_dataset)
    if not source_dataset_metadata:
        raise IncorrectUsageError(f"The source dataset with name '{source_dataset}' not found")

    request = CopyQualityStandardRequest(
        dataset_metadata.id,
        source_dataset_metadata.id,
        include_metric_groups=include_metric_groups,
        include_test_cases=include_test_cases,
    )

    response = krequests.put(
        Path.COPY_FROM_DATASET,
        json=asdict(request),
        api_version="v2",
    )
    krequests.raise_for_status(response)

    metric_groups = response.json().get("metric_groups", [])
    test_cases = response.json().get("stratifications", [])
    return metric_groups, test_cases



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import field
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from kolena._api.v1.batched_load import BatchedLoad
from kolena._utils.pydantic_v1.dataclasses import dataclass


class Path(str, Enum):
    UPLOAD_RESULTS = "/model/upload-results"
    LOAD_RESULTS = "/model/load-results"
    LOAD_BY_TAG = "/model/load-by-tag"
    LOAD_EVAL_CONFIG = "/model/load-eval-config"
    LOAD_BY_NAME = "/model/load"
    LOAD_BY_DATASET = "/model/load-by-dataset"


@dataclass(frozen=True)
class LoadByNameRequest:
    name: str


@dataclass(frozen=True)
class IdRequest:
    id: int


@dataclass(frozen=True)
class LoadResultsRequest(BatchedLoad.BaseInitDownloadRequest):
    model: str
    dataset: str
    commit: Optional[str] = None
    include_extracted_properties: bool = False


@dataclass(frozen=True)
class UploadResultsRequest:
    model: str
    uuid: str
    dataset_id: int
    sources: Optional[List[Dict[str, str]]]
    tags: List[str] = field(default_factory=list)


@dataclass(frozen=True)
class UploadResultsResponse:
    n_inserted: int
    n_updated: int
    model_id: int
    eval_config_id: Union[int, List[int]]


@dataclass(frozen=True)
class EntityData:
    id: int
    name: str


@dataclass(frozen=True)
class LoadByTagRequest:
    tag: str


@dataclass(frozen=True)
class LoadByTagResponse:
    models: List[EntityData]


@dataclass(frozen=True)
class LoadByDatasetRequest:
    dataset_id: int


@dataclass(frozen=True)
class LoadEvalConfigRequest:
    id: Optional[int] = None
    data: Optional[Dict[str, Any]] = None


@dataclass(frozen=True)
class EvalConfigEntityData:
    id: int
    data: Optional[Dict[str, Any]]


@dataclass(frozen=True)
class ModelWithEvalConfig:
    model_id: int
    eval_config_id: int



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from pydantic import Field
from pydantic import field_validator
from pydantic.dataclasses import dataclass
from typing_extensions import Annotated

MetricFormat = Literal["default", "integer", "decimal", "percentage", "scientific", "dollars", "euros"]


@dataclass(frozen=True, order=True)
class Highlight:
    higherIsBetter: Union[bool, None] = None


@dataclass(frozen=True)
class BaseMetric:
    label: str
    source: Literal["result"]


AveragingMethod = Literal["macro", "micro", "weighted"]
ClassifierAggregator = Literal["precision", "recall", "f1_score"]
SimpleAggregator = Literal["count", "min", "max", "mean", "median", "stddev", "sum"]
Source = Literal["datapoint", "result"]


@dataclass(frozen=True)
class SimpleParams:
    key: str


@dataclass(frozen=True)
class SimpleMetric(BaseMetric):
    aggregator: SimpleAggregator
    params: SimpleParams
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True)
class CategoricalAggregator:
    aggregator: Literal["count", "rate"]
    fieldValue: Union[bool, str]


@dataclass(frozen=True)
class CategoricalMetric(BaseMetric):
    aggregator: CategoricalAggregator
    params: SimpleParams
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True, order=True)
class BinaryClassificationAggregator:
    aggregator: Union[Literal["accuracy"], Literal["auc"], ClassifierAggregator]
    type: Literal["binary-classification"]


@dataclass(frozen=True, order=True)
class MulticlassClassificationAggregator:
    aggregator: Union[Literal["accuracy"], ClassifierAggregator]
    averagingMethod: Annotated[Union[AveragingMethod, None], Field(validate_default=True)] = None

    @field_validator("averagingMethod")
    @classmethod
    def set_null_averaging_method(cls, v: Union[str, None]) -> str:
        return v or "macro"


@dataclass(frozen=True, order=True)
class RegressionAggregator:
    aggregator: Literal["mae", "mse", "rmse", "r_squared", "pearson_corr", "spearman_corr"]


@dataclass(frozen=True, order=True)
class ClassificationParams:
    groundTruthField: str
    inferenceField: str
    threshold: Union[int, float, None] = None
    include_auc: bool = False


@dataclass(frozen=True)
class ClassificationMetric(BaseMetric):
    aggregator: Union[BinaryClassificationAggregator, MulticlassClassificationAggregator, RegressionAggregator]
    params: ClassificationParams
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True, order=True)
class ObjectDetectionAggregator:
    aggregator: Union[Literal["average_precision"], ClassifierAggregator]
    averagingMethod: Annotated[Union[AveragingMethod, None], Field(validate_default=True)] = None

    @field_validator("averagingMethod")
    @classmethod
    def set_null_averaging_method(cls, v: Union[str, None]) -> str:
        return v or "macro"


@dataclass(frozen=True, order=True)
class ObjectDetectionParams:
    fnField: str
    fpField: str
    tpField: str
    scoreField: str
    labelField: Union[str, None] = None


@dataclass(frozen=True)
class ObjectDetectionMetric(BaseMetric):
    aggregator: ObjectDetectionAggregator
    params: ObjectDetectionParams
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True, order=True)
class ThresholdedObjectPoweredAggregator:
    aggregator: Union[Literal["average_precision"], ClassifierAggregator]
    averagingMethod: AveragingMethod = "macro"
    type: Literal["thresholded-object-powered"] = "thresholded-object-powered"


@dataclass(frozen=True, order=True)
class ThresholdedObjectPoweredParams:
    threshold: float
    thresholded_object_field: str
    precision_tp_field: str
    precision_fp_field: str
    recall_tp_field: str
    recall_fn_field: str
    label: Optional[str] = None


@dataclass(frozen=True)
class ThresholdedObjectPoweredMetric(BaseMetric):
    aggregator: ThresholdedObjectPoweredAggregator
    params: ThresholdedObjectPoweredParams
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True)
class PrecisionParams:
    fp: str
    tp: str


@dataclass(frozen=True)
class PrecisionMetric(BaseMetric):
    aggregator: Literal["precision"]
    params: PrecisionParams
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True)
class RecallParams:
    fn: str
    tp: str


@dataclass(frozen=True)
class RecallMetric(BaseMetric):
    aggregator: Literal["recall"]
    params: RecallParams
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True)
class F1Params:
    fn: str
    fp: str
    tp: str


@dataclass(frozen=True)
class F1Metric(BaseMetric):
    aggregator: Literal["f1_score"]
    params: F1Params
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True)
class AccuracyParams:
    fp: str
    fn: str
    tp: str
    tn: str


@dataclass(frozen=True)
class AccuracyMetric(BaseMetric):
    aggregator: Literal["accuracy"]
    params: AccuracyParams
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True)
class RateParams:
    fp: str
    tn: str


@dataclass(frozen=True)
class RateMetric(BaseMetric):
    aggregator: Literal["false_positive_rate", "true_negative_rate"]
    params: RateParams
    format: Union[MetricFormat, None] = None
    highlight: Union[Highlight, None] = None


@dataclass(frozen=True)
class WeightedMetricAggregator:
    type: Literal["weighted-metric"] = "weighted-metric"
    aggregator: Literal["mean"] = "mean"


@dataclass(frozen=True)
class WeightedMetricParams:
    metric_field: str
    weight_field: str
    # TODO: mark these as mandatory when marina supports non-result sources
    metric_source: Source = "result"
    weight_source: Source = "result"


@dataclass(frozen=True)
class WeightedMetric(BaseMetric):
    aggregator: WeightedMetricAggregator
    params: WeightedMetricParams
    highlight: Union[Highlight, None] = None


FormulaMetric = Union[
    SimpleMetric,
    CategoricalMetric,
    PrecisionMetric,
    RecallMetric,
    F1Metric,
    AccuracyMetric,
    RateMetric,
    WeightedMetric,
]

TaskMetric = Union[ClassificationMetric, ObjectDetectionMetric, ThresholdedObjectPoweredMetric]

Metric = Union[
    FormulaMetric,
    TaskMetric,
]


@dataclass(frozen=True)
class MetricGroup:
    name: str
    metrics: List[Metric]

    @field_validator("metrics")
    @classmethod
    def metric_label_unique(cls, metrics: List[Metric]) -> List[Metric]:
        if len(metrics) > len({metric.label for metric in metrics}):
            raise ValueError("Metric labels must be unique.")
        return metrics

    def __hash__(self) -> int:
        return hash(
            (
                self.name,
                tuple(self.metrics),
            ),
        )


@dataclass(frozen=True)
class DatasetMetricGroups:
    dataset_id: int
    dataset_name: str
    metric_groups: List[MetricGroup]



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import field
from typing import Dict
from typing import List
from typing import Literal
from typing import Union

from pydantic import StrictBool
from pydantic import StrictStr
from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class CategoricalBucket:
    value: Union[StrictStr, StrictBool, int, None]
    count: int


@dataclass(frozen=True)
class NumericBucket:
    lower: float
    upper: float
    count: int
    interval_type: Literal["closed", "right-open"] = "right-open"


@dataclass(frozen=True)
class CategoricalStats:
    histogram: List[CategoricalBucket]
    data_type: Literal["categorical"] = "categorical"


@dataclass(frozen=True)
class NumericStats:
    count: int
    mean: float
    median: float
    min: float
    max: float
    stddev: float
    sum: float
    histogram: List[NumericBucket]
    data_type: Literal["numeric"] = "numeric"


@dataclass(frozen=True)
class ArraySizeStats:
    mean: float
    median: float
    min: int
    max: int
    stddev: float
    sum: int
    histogram: List[CategoricalBucket]


@dataclass(frozen=True)
class ArrayStats:
    length: ArraySizeStats
    data_type: Literal["array"] = "array"


@dataclass(frozen=True)
class FieldStats:
    key: str
    data: Union[CategoricalStats, NumericStats, ArrayStats]


@dataclass(frozen=True)
class SingleStatsResponse:
    datapoint: List[FieldStats] = field(default_factory=list)
    # key by model id first, then eval_config_id
    result: Dict[str, Dict[str, List[FieldStats]]] = field(default_factory=dict)
    llm: List[FieldStats] = field(default_factory=list)


@dataclass(frozen=True)
class FieldStatsResponse:
    stats: List[SingleStatsResponse]



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kolena._utils.pydantic_v1 import validate_arguments
from kolena._utils.validators import ValidatorConfig


@validate_arguments(config=ValidatorConfig)
def validate_label(label: str) -> None:
    if label.strip() == "":
        raise ValueError("label must contain non-whitespace characters", label)


@validate_arguments(config=ValidatorConfig)
def validate_confidence(confidence: float) -> None:
    if not (0 <= confidence <= 1):
        raise ValueError("confidence must be between 0 and 1 (inclusive)", confidence)



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json

from kolena._api.v1.repository import CreateRepositoryRequest
from kolena._api.v1.repository import Path
from kolena._utils import krequests


def create(repository: str) -> None:
    response = krequests.post(
        endpoint_path=Path.CREATE.value,
        data=json.dumps(dataclasses.asdict(CreateRepositoryRequest(repository=repository))),
    )
    krequests.raise_for_status(response)



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum

KOLENA_TOKEN_ENV = "KOLENA_TOKEN"


class BatchSize(int, Enum):
    UPLOAD_RECORDS = 10_000_000
    UPLOAD_EMBEDDINGS = 1_000_000

    LOAD_RECORDS = UPLOAD_RECORDS
    LOAD_SAMPLES = 1_000_000


class FieldName(str, Enum):
    TEST_CASE_NAME = "Test Case name"
    TEST_SUITE_NAME = "Test Suite name"
    MODEL_NAME = "Model name"



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io

import numpy as np
from PIL import Image

from kolena.errors import InputValidationError


def encode_png(image: np.ndarray, mode: str) -> io.BytesIO:
    """
    Encodes an image into an in-memory PNG file that is represented as binary data. It is used when you want to upload
    a 2 or 3-dimensional image in a NumPy array format to cloud.

    It can be used in conjunction with
    [`colorized_activation_map`][kolena.workflow.visualization.colorize_activation_map] when uploading an
    activation map.

    :param image: A 2D or 3D NumPy array, shaped either `(h, w)`, `(h, w, 1)`, `(h, w, 3)`, or `(h, w, 4)`
    :param mode: A [PIL mode](https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes)
    :return: The in-memory PNG file represented as binary data.
    """
    if image.size == 0:
        raise InputValidationError("input array is empty")

    if len(image.shape) < 2 or len(image.shape) > 3:
        raise InputValidationError(
            f"input array must have 2 or 3 dimensions, but received {len(image.shape)}\n"
            "Here are supported input shape: (h, w), (h, w, 1), (h, w, 3), or (h, w, 4)",
        )

    pil_image = Image.fromarray(image, mode=mode)
    image_buf = io.BytesIO()
    pil_image.save(image_buf, format="png")
    image_buf.seek(0)

    return image_buf



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import math
import re
import warnings
from collections.abc import Callable
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import numpy as np
import pandas as pd


def df_apply(df: pd.DataFrame, func: Callable, **kwargs: Any) -> pd.DataFrame:
    """
    pandas.DataFrame.apply wrapper
    """
    default_kwargs = dict(convert_dtype=False)
    with warnings.catch_warnings():
        # Silence the warning for now -- pandas devs have deprecated this without providing a good alternative:
        # https://github.com/pandas-dev/pandas/pull/52257#issuecomment-1684888371
        warnings.simplefilter(action="ignore", category=FutureWarning)
        return df.apply(func, **default_kwargs, **kwargs)


def _flatten_dict(d: Dict[str, Any], parent_key: str = "", sep: str = ".", max_level: int = 0) -> Dict[str, Any]:
    items: List[Tuple] = []
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict) and max_level > 0:
            items.extend(_flatten_dict(v, new_key, sep, max_level - 1).items())
        else:
            items.append((new_key, v))
    return dict(items)


def _normalize_dicts(dict_list: List[Dict[str, Any]], sep: str = ".", max_level: int = 0) -> List[Dict[str, Any]]:
    # Flatten all dictionaries in the list
    flattened_dicts = [_flatten_dict(d, sep=sep, max_level=max_level) for d in dict_list]

    # Get all unique keys across all dictionaries
    all_keys = set().union(*flattened_dicts)

    # Normalize all dictionaries to have all keys, filling missing ones with None
    for d in flattened_dicts:
        for key in all_keys:
            if key not in d:
                d[key] = None

    return flattened_dicts


def _try_parse(value: Any) -> Any:
    """
    Try parsing the input value into a more appropriate Python data type.
    This function handles the following conversions:

    - Empty strings are converted to None.
    - Strings that represent numbers are converted to numerical types (int or float).
    - JSON strings are converted to corresponding Python objects (lists or dictionaries).
    - Strings "true" and "false" are converted to their boolean equivalents.
    - NumPy arrays are converted to Python lists.
    - NaN values are converted to None.

    If none of the above conversions are applicable, the original value is returned.

    Parameters:
    value (Any): The input value to be parsed.

    Returns:
    Any: The parsed value if a conversion was successful, otherwise the original value.
    """

    # Handle NaN values
    if isinstance(value, float) and math.isnan(value):
        return None

    # Match behaviour: json.loads("NaN") == float('nan')
    if isinstance(value, str) and value == "NaN":
        return None

    # Convert empty strings to None
    if isinstance(value, str) and value == "":
        return None

    # Normalize NumPy arrays into Python lists to avoid downstream type errors
    if isinstance(value, np.ndarray):
        return value.tolist()

    # If the value is not a string, return it as is
    if not isinstance(value, str):
        return value

    # Convert strings "true" and "false" to boolean values
    if value.lower() in ["true", "false"]:
        return value.lower() == "true"

    # some string id could look like a scientific notation with e
    scientific_notation_pattern = re.compile(r"[+-]?\d+(\.\d+)?[eE][+-]?\d+")
    if scientific_notation_pattern.match(value):
        return value

    # Attempt to convert the string to a numeric type
    try:
        value_numeric = pd.to_numeric(value, errors="coerce")
        if not np.isnan(value_numeric):
            return value_numeric
    except Exception:
        pass

    # Only attempt to parse JSON arrays or objects
    if not (value.startswith("[") or value.startswith("{")):
        if value == "" or value == "null":
            return None
        elif value.startswith(("'", '"')) and value.endswith(("'", '"')):
            value = value[1:-1]
        return value

    # Attempt to parse the string as JSON
    try:
        return json.loads(value)
    except Exception:
        return value


def json_normalize(data: List[Dict[str, Any]], max_level: int = 0) -> pd.DataFrame:
    """
    hand roll pandas.json_normalize implementation
    use None for missing value
    """
    normalized_data = _normalize_dicts(data, max_level=max_level)
    return pd.DataFrame.from_dict(normalized_data, dtype=object)



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict
from typing import Literal
from typing import Union

import pandas as pd

from kolena._experimental.object_detection.dataset import upload_object_detection_results


def upload_instance_segmentation_results(
    dataset_name: str,
    model_name: str,
    df: pd.DataFrame,
    *,
    ground_truths_field: str = "ground_truths",
    raw_inferences_field: str = "raw_inferences",
    iou_threshold: float = 0.5,
    threshold_strategy: Union[Literal["F1-Optimal"], float, Dict[str, float]] = "F1-Optimal",
    min_confidence_score: float = 0.01,
    batch_size: int = 10_000,
) -> None:
    """
    Compute metrics and upload results of the model for the dataset.

    Dataframe `df` should include a `locator` column that would match to that of corresponding datapoint and
    an `inference` column that should be a list of scored [`Polygons`][kolena.annotation.Polygon].

    :param dataset_name: Dataset name.
    :param model_name: Model name.
    :param df: Dataframe for model results.
    :param ground_truths_field: Field name in datapoint with ground truth polygons, defaulting to `"ground_truths"`.
    :param raw_inferences_field: Column in model result DataFrame with raw inference polygons,
    defaulting to `"raw_inferences"`.
    :param iou_threshold: The [IoU ↗](../../metrics/iou.md) threshold, defaulting to `0.5`.
    :param threshold_strategy: The confidence threshold strategy. It can either be a fixed confidence threshold such
        as `0.5` or `0.75`, or `"F1-Optimal"` to find the threshold maximizing F1 score..
    :param min_confidence_score: The minimum confidence score to consider for the evaluation. This is usually set to
        reduce noise by excluding inferences with low confidence score.
    :return:
    """
    return upload_object_detection_results(
        dataset_name,
        model_name,
        df,
        ground_truths_field=ground_truths_field,
        raw_inferences_field=raw_inferences_field,
        iou_threshold=iou_threshold,
        threshold_strategy=threshold_strategy,
        min_confidence_score=min_confidence_score,
        batch_size=batch_size,
    )



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from sys import float_info
from typing import Optional

"""
 The following Python functions for calculating normal and
    chi-square probabilities and critical values were adapted by
    John Walker's Javascript implementation from C implementations
    written by Gary Perlman of Wang Institute, Tyngsboro, MA
    01879.  Both the original C code and this JavaScript edition
    are in the public domain.  */

/*  POZ  --  probability of normal z value
    Adapted from a polynomial approximation in:
            Ibbetson D, Algorithm 209
            Collected Algorithms of the CACM 1963 p. 616
    Note:
            This routine has six digit accuracy, so it is only useful for absolute
            z values <= 6.  For z values > to 6.0, poz() returns 0.0.
"""


def z_score_to_probability(z: float) -> float:
    Z_MAX = 6.0
    if z == 0.0:
        x = 0.0
    else:
        y = 0.5 * abs(z)
        if y > Z_MAX * 0.5:
            x = 1.0
        elif y < 1.0:
            w = y * y
            x = (
                (
                    (
                        (
                            (
                                (
                                    (((0.000124818987 * w - 0.001075204047) * w + 0.005198775019) * w - 0.019198292004)
                                    * w
                                    + 0.059054035642
                                )
                                * w
                                - 0.151968751364
                            )
                            * w
                            + 0.319152932694
                        )
                        * w
                        - 0.531923007300
                    )
                    * w
                    + 0.797884560593
                )
                * y
                * 2.0
            )
        else:
            y -= 2.0
            x = (
                (
                    (
                        (
                            (
                                (
                                    (
                                        (
                                            (
                                                (
                                                    (
                                                        ((-0.000045255659 * y + 0.000152529290) * y - 0.000019538132)
                                                        * y
                                                        - 0.000676904986
                                                    )
                                                    * y
                                                    + 0.001390604284
                                                )
                                                * y
                                                - 0.000794620820
                                            )
                                            * y
                                            - 0.002034254874
                                        )
                                        * y
                                        + 0.006549791214
                                    )
                                    * y
                                    - 0.010557625006
                                )
                                * y
                                + 0.011630447319
                            )
                            * y
                            - 0.009279453341
                        )
                        * y
                        + 0.005353579108
                    )
                    * y
                    - 0.002141268741
                )
                * y
                + 0.000535310849
            ) * y + 0.999936657524
    if z > 0.0:
        return (x + 1.0) * 0.5
    else:
        return (1.0 - x) * 0.5


def probability_to_z_score(p: float) -> float:
    Z_MAX = 6.0
    Z_EPSILON = 0.000001  # Accuracy of z approximation
    minz = -Z_MAX
    maxz = Z_MAX
    zval = 0.0
    pp = p
    if pp < 0.0:
        pp = 0.0
    if pp > 1.0:
        pp = 1.0

    while maxz - minz > Z_EPSILON:
        pval = z_score_to_probability(zval)
        if pval > pp:
            maxz = zval
        else:
            minz = zval
        zval = (maxz + minz) * 0.5
    return zval


def confidence_level_to_z_score(confidence_level: float) -> float:
    if confidence_level < 0 or confidence_level >= 1:
        raise ValueError(f"Invalid confidence level: value '{confidence_level}' must be in [0, 1)")
    probability = (1 + confidence_level) / 2.0
    return probability_to_z_score(probability)


def margin_of_error(n_samples: int, confidence_level: float = 0.95, positive_sample_rate: float = 0.5) -> float:
    if n_samples <= 0:
        return 0.0
    z_score = confidence_level_to_z_score(confidence_level)
    margin_error = z_score * math.sqrt((positive_sample_rate * (1 - positive_sample_rate)) / n_samples)
    return margin_error * 100  # percentage


def round_up_to_nearest_power_of_10(value: float) -> int:
    return 10 ** math.ceil(math.log10(value))


def get_delta_percentage(value_a: float, value_b: float, max_value: Optional[int] = None) -> float:
    upper_limit = None

    if max_value is not None and max_value > 0:
        upper_limit = round_up_to_nearest_power_of_10(max_value)
        if upper_limit >= 1000:
            upper_limit = max_value
        if math.isnan(upper_limit) or not math.isfinite(upper_limit) or upper_limit == 0:
            upper_limit = None

    denominator = upper_limit if upper_limit is not None else value_a + float_info.epsilon
    # Adding sys.float_info.epsilon to value_a to prevent division by zero if value_a is zero.

    delta_percentage = (100 * (value_a - value_b)) / denominator
    return delta_percentage


QUANTILES = {
    1: "all",
    2: "half",
    3: "tertile",
    4: "quartile",
    5: "quintile",
    6: "sextile",
    7: "septile",
    8: "octile",
    10: "decile",
    100: "percentile",
}


def get_quantile(n_buckets: int) -> str:
    return QUANTILES.get(n_buckets, "quantile")


def ordinal(n: int) -> str:
    if 11 <= (n % 100) <= 13:
        suffix = "th"
    else:
        suffix = ["th", "st", "nd", "rd", "th"][min(n % 10, 4)]
    return str(n) + suffix



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# mypy: ignore-errors
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from kolena._experimental.instance_segmentation.workflow import EvaluatorConfiguration
from kolena._experimental.instance_segmentation.workflow import GroundTruth
from kolena._experimental.instance_segmentation.workflow import Inference
from kolena._experimental.instance_segmentation.workflow import TestCase
from kolena._experimental.instance_segmentation.workflow import TestSample
from kolena._experimental.instance_segmentation.workflow import TestSuite
from kolena._experimental.object_detection.evaluator import ObjectDetectionEvaluator
from kolena._experimental.object_detection.workflow import GroundTruth as ObjectGroundTruth
from kolena._experimental.object_detection.workflow import Inference as ObjectInference
from kolena._experimental.object_detection.workflow import MetricsTestCase
from kolena._experimental.object_detection.workflow import MetricsTestSample
from kolena._experimental.object_detection.workflow import MetricsTestSuite
from kolena._experimental.object_detection.workflow import TestSampleMetrics
from kolena._experimental.object_detection.workflow import TestSampleMetricsSingleClass
from kolena.workflow import Evaluator
from kolena.workflow import Plot


class InstanceSegmentationEvaluator(Evaluator):
    """
    This `InstanceSegmentationEvaluator` transforms inferences into metrics for the instance segmentation workflow for a
    single class or multiple classes.

    When a [`EvaluatorConfiguration`][kolena._experimental.instance_segmentation.workflow.EvaluatorConfiguration] is
    configured to use an F1-Optimal threshold strategy, the evaluator requires that the first test case retrieved for
    a test suite contains the complete sample set.

    For additional functionality, see the associated [base class documentation][kolena.workflow.evaluator.Evaluator].
    """

    def __init__(self, configurations: Optional[List[EvaluatorConfiguration]] = None):
        self.evaluator = ObjectDetectionEvaluator()
        super().__init__(configurations)

    def compute_test_sample_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> List[Tuple[TestSample, Union[TestSampleMetrics, TestSampleMetricsSingleClass]]]:
        inferences = InstanceSegmentationEvaluator._convert_inferences(inferences)
        return self.evaluator.compute_test_sample_metrics(test_case, inferences, configuration)

    def compute_test_case_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[MetricsTestSample],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> MetricsTestCase:
        inferences = InstanceSegmentationEvaluator._convert_inferences(inferences)
        return self.evaluator.compute_test_case_metrics(
            test_case,
            inferences,
            metrics,
            configuration,
        )

    def compute_test_case_plots(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[Union[TestSampleMetrics, TestSampleMetricsSingleClass]],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> Optional[List[Plot]]:
        inferences = InstanceSegmentationEvaluator._convert_inferences(inferences)
        return self.evaluator.compute_test_case_plots(
            test_case,
            inferences,
            metrics,
            configuration,
        )

    def compute_test_suite_metrics(
        self,
        test_suite: TestSuite,
        metrics: List[Tuple[TestCase, MetricsTestCase]],
        configuration: Optional[EvaluatorConfiguration] = None,
    ) -> Optional[MetricsTestSuite]:
        return self.evaluator.compute_test_suite_metrics(test_suite, metrics, configuration)

    @staticmethod
    def _convert_inferences(
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
    ) -> List[Tuple[TestSample, ObjectGroundTruth, ObjectInference]]:
        return [
            (
                ts,
                ObjectGroundTruth(bboxes=gt.polygons, ignored_bboxes=gt.ignored_polygons),
                ObjectInference(bboxes=inf.polygons, ignored=inf.ignored),
            )
            for ts, gt, inf in inferences
        ]



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# mypy: disable-error-code="override"
from collections import defaultdict
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import numpy as np

from kolena._experimental.object_detection import ClassMetricsPerTestCase
from kolena._experimental.object_detection import GroundTruth
from kolena._experimental.object_detection import Inference
from kolena._experimental.object_detection import TestCase
from kolena._experimental.object_detection import TestCaseMetrics
from kolena._experimental.object_detection import TestSample
from kolena._experimental.object_detection import TestSampleMetrics
from kolena._experimental.object_detection import TestSuite
from kolena._experimental.object_detection import TestSuiteMetrics
from kolena._experimental.object_detection import ThresholdConfiguration
from kolena._experimental.object_detection.utils import compute_average_precision
from kolena._experimental.object_detection.utils import compute_confusion_matrix_plot
from kolena._experimental.object_detection.utils import compute_f1_plot_multiclass
from kolena._experimental.object_detection.utils import compute_optimal_f1_threshold_multiclass
from kolena._experimental.object_detection.utils import compute_pr_curve
from kolena._experimental.object_detection.utils import compute_pr_plot_multiclass
from kolena._experimental.object_detection.utils import filter_inferences
from kolena.workflow import Evaluator
from kolena.workflow import EvaluatorConfiguration
from kolena.workflow import Plot
from kolena.workflow.annotation import ScoredLabel
from kolena.workflow.metrics import f1_score as compute_f1_score
from kolena.workflow.metrics import match_inferences_multiclass
from kolena.workflow.metrics import MulticlassInferenceMatches
from kolena.workflow.metrics import precision as compute_precision
from kolena.workflow.metrics import recall as compute_recall


class MulticlassObjectDetectionEvaluator(Evaluator):
    """
    The `MulticlassObjectDetectionEvaluator` transforms inferences into metrics for the object detection workflow for
    multiple classes.

    When a [`ThresholdConfiguration`][kolena._experimental.object_detection.workflow.ThresholdConfiguration] is
    configured to use an F1-Optimal threshold strategy, the evaluator requires that the first test case retrieved for
    a test suite contains the complete sample set.

    For additional functionality, see the associated [base class documentation][kolena.workflow.evaluator.Evaluator].
    """

    threshold_cache: Dict[str, Dict[str, float]]  # configuration -> label -> threshold
    """
    Assumes that the first test case retrieved for the test suite contains the complete sample set to be used for
    F1-Optimal threshold computation. Subsequent requests for a given threshold strategy (for other test cases) will
    hit this cache and use the previously computed population level confidence thresholds.
    """

    locators_by_test_case: Dict[str, List[str]]
    """
    Keeps track of test sample locators for each test case (used for total # of image count in aggregated metrics).
    """

    matchings_by_test_case: Dict[str, Dict[str, List[MulticlassInferenceMatches]]]
    """
    Caches matchings per configuration and test case for faster test case metric and plot computation.
    """

    def __init__(self, configurations: Optional[List[EvaluatorConfiguration]] = None):
        super().__init__(configurations)
        self.threshold_cache = {}
        self.locators_by_test_case = {}
        self.matchings_by_test_case = defaultdict(lambda: defaultdict(list))

    def test_sample_metrics_ignored(
        self,
    ) -> TestSampleMetrics:
        return TestSampleMetrics(
            TP=[],
            FP=[],
            FN=[],
            Confused=[],
            count_TP=0,
            count_FP=0,
            count_FN=0,
            count_Confused=0,
            has_TP=False,
            has_FP=False,
            has_FN=False,
            has_Confused=False,
            ignored=True,
            max_confidence_above_t=None,
            min_confidence_above_t=None,
            thresholds=[],
        )

    def test_sample_metrics(
        self,
        bbox_matches: MulticlassInferenceMatches,
        thresholds: Dict[str, float],
    ) -> TestSampleMetrics:
        tp = [inf for _, inf in bbox_matches.matched if inf.score >= thresholds[inf.label]]
        fp = [inf for inf in bbox_matches.unmatched_inf if inf.score >= thresholds[inf.label]]
        fn = [gt for gt, _ in bbox_matches.unmatched_gt] + [
            gt for gt, inf in bbox_matches.matched if inf.score < thresholds[inf.label]
        ]
        confused = [
            inf for _, inf in bbox_matches.unmatched_gt if inf is not None and inf.score >= thresholds[inf.label]
        ]
        non_ignored_inferences = tp + fp
        scores = [inf.score for inf in non_ignored_inferences]
        inference_labels = {inf.label for _, inf in bbox_matches.matched} | {
            inf.label for inf in bbox_matches.unmatched_inf
        }
        fields = [
            ScoredLabel(label=label, score=thresholds[label])
            for label in sorted(thresholds.keys())
            if label in inference_labels
        ]
        return TestSampleMetrics(
            TP=tp,
            FP=fp,
            FN=fn,
            Confused=confused,
            count_TP=len(tp),
            count_FP=len(fp),
            count_FN=len(fn),
            count_Confused=len(confused),
            has_TP=len(tp) > 0,
            has_FP=len(fp) > 0,
            has_FN=len(fn) > 0,
            has_Confused=len(confused) > 0,
            ignored=False,
            max_confidence_above_t=max(scores) if len(scores) > 0 else None,
            min_confidence_above_t=min(scores) if len(scores) > 0 else None,
            thresholds=fields,
        )

    def compute_image_metrics(
        self,
        ground_truth: GroundTruth,
        inference: Inference,
        configuration: ThresholdConfiguration,
        test_case_name: str,
    ) -> TestSampleMetrics:
        assert configuration is not None, "must specify configuration"
        thresholds = self.get_confidence_thresholds(configuration)
        if inference.ignored:
            return self.test_sample_metrics_ignored()

        filtered_inferences = filter_inferences(
            inferences=inference.bboxes,
            confidence_score=configuration.min_confidence_score,
        )
        bbox_matches: MulticlassInferenceMatches = match_inferences_multiclass(
            ground_truth.bboxes,
            filtered_inferences,
            ignored_ground_truths=ground_truth.ignored_bboxes,
            mode="pascal",
            iou_threshold=configuration.iou_threshold,
        )
        self.matchings_by_test_case[configuration.display_name()][test_case_name].append(bbox_matches)

        return self.test_sample_metrics(bbox_matches, thresholds)

    def compute_and_cache_f1_optimal_thresholds(
        self,
        configuration: ThresholdConfiguration,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
    ) -> None:
        if configuration.threshold_strategy != "F1-Optimal":
            return

        if configuration.display_name() in self.threshold_cache.keys():
            return

        all_bbox_matches = [
            match_inferences_multiclass(
                ground_truth.bboxes,
                filter_inferences(inferences=inference.bboxes, confidence_score=configuration.min_confidence_score),
                ignored_ground_truths=ground_truth.ignored_bboxes,
                mode="pascal",
                iou_threshold=configuration.iou_threshold,
            )
            for _, ground_truth, inference in inferences
            if not inference.ignored
        ]
        optimal_thresholds = compute_optimal_f1_threshold_multiclass(all_bbox_matches)
        self.threshold_cache[configuration.display_name()] = defaultdict(
            lambda: configuration.min_confidence_score,
            optimal_thresholds,
        )

    def compute_test_sample_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],  # type: ignore
        configuration: Optional[ThresholdConfiguration] = None,  # type: ignore
    ) -> List[Tuple[TestSample, TestSampleMetrics]]:
        assert configuration is not None, "must specify configuration"
        # compute thresholds to cache values for subsequent steps
        self.compute_and_cache_f1_optimal_thresholds(configuration, inferences)
        return [
            (ts, self.compute_image_metrics(gt, inf, configuration, test_case.name))  # type: ignore
            for ts, gt, inf in inferences
        ]

    def bbox_matches_and_count_for_one_label(
        self,
        matchings: List[MulticlassInferenceMatches],
        label: str,
    ) -> Tuple[MulticlassInferenceMatches, int]:
        match_matched = []
        match_unmatched_gt = []
        match_unmatched_inf = []
        samples_count = 0
        # filter the matching to only consider one class
        for match in matchings:
            sample_flag = False
            for gt, inf in match.matched:
                if gt.label == label:
                    sample_flag = True
                    match_matched.append((gt, inf))
            for gt, inf in match.unmatched_gt:
                if gt.label == label:
                    sample_flag = True
                    match_unmatched_gt.append((gt, inf))
            for inf in match.unmatched_inf:
                if inf.label == label:
                    match_unmatched_inf.append(inf)
            if sample_flag:
                samples_count += 1

        bbox_matches_for_one_label = MulticlassInferenceMatches(
            matched=match_matched,
            unmatched_gt=match_unmatched_gt,
            unmatched_inf=match_unmatched_inf,
        )

        return bbox_matches_for_one_label, samples_count

    def class_metrics_per_test_case(
        self,
        label: str,
        thresholds: Dict[str, float],
        class_matches: MulticlassInferenceMatches,
        samples_count: int,
        average_precision: float,
    ) -> ClassMetricsPerTestCase:
        matched = class_matches.matched
        unmatched_gt = class_matches.unmatched_gt
        unmatched_inf = class_matches.unmatched_inf

        tp = [inf for _, inf in matched if inf.score >= thresholds[inf.label]]
        fp = [inf for inf in unmatched_inf if inf.score >= thresholds[inf.label]]
        fn = [gt for gt, _ in unmatched_gt] + [gt for gt, inf in matched if inf.score < thresholds[inf.label]]
        tp_count = len(tp)
        fp_count = len(fp)
        fn_count = len(fn)
        precision = compute_precision(tp_count, fp_count)
        recall = compute_recall(tp_count, fn_count)
        f1_score = compute_f1_score(tp_count, fp_count, fn_count)

        return ClassMetricsPerTestCase(
            Class=label,
            nImages=samples_count,
            Threshold=thresholds[label],
            Objects=tp_count + fn_count,
            Inferences=tp_count + fp_count,
            TP=tp_count,
            FN=fn_count,
            FP=fp_count,
            Precision=precision,
            Recall=recall,
            F1=f1_score,
            AP=average_precision,
        )

    def compute_aggregate_label_metrics(
        self,
        matchings: List[MulticlassInferenceMatches],
        label: str,
        thresholds: Dict[str, float],
    ) -> ClassMetricsPerTestCase:
        class_matches, samples_count = self.bbox_matches_and_count_for_one_label(matchings, label)

        average_precision = 0.0
        baseline_pr_curve = compute_pr_curve([class_matches])
        if baseline_pr_curve is not None:
            average_precision = compute_average_precision(baseline_pr_curve.y, baseline_pr_curve.x)  # type: ignore

        return self.class_metrics_per_test_case(label, thresholds, class_matches, samples_count, average_precision)

    def test_case_metrics(
        self,
        per_class_metrics: List[ClassMetricsPerTestCase],
        metrics: List[TestSampleMetrics],
    ) -> TestCaseMetrics:
        tp_count = sum(im.count_TP for im in metrics)
        fp_count = sum(im.count_FP for im in metrics)
        fn_count = sum(im.count_FN for im in metrics)
        ignored_count = sum(1 if im.ignored else 0 for im in metrics)
        macro_prec_data = np.mean([data.Precision for data in per_class_metrics])
        macro_rec_data = np.mean([data.Recall for data in per_class_metrics])
        return TestCaseMetrics(
            PerClass=per_class_metrics,
            Objects=tp_count + fn_count,
            Inferences=tp_count + fp_count,
            TP=tp_count,
            FN=fn_count,
            FP=fp_count,
            nIgnored=ignored_count,
            macro_Precision=macro_prec_data if per_class_metrics else 0.0,  # type: ignore
            macro_Recall=macro_rec_data if per_class_metrics else 0.0,  # type: ignore
            macro_F1=np.mean([data.F1 for data in per_class_metrics]) if per_class_metrics else 0.0,  # type: ignore
            mean_AP=np.mean([data.AP for data in per_class_metrics]) if per_class_metrics else 0.0,  # type: ignore
            micro_Precision=compute_precision(tp_count, fp_count),
            micro_Recall=compute_recall(tp_count, fn_count),
            micro_F1=compute_f1_score(tp_count, fp_count, fn_count),
        )

    def compute_test_case_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],  # type: ignore
        metrics: List[TestSampleMetrics],  # type: ignore
        configuration: Optional[ThresholdConfiguration] = None,  # type: ignore
    ) -> TestCaseMetrics:
        assert configuration is not None, "must specify configuration"
        thresholds = self.get_confidence_thresholds(configuration)
        all_bbox_matches = self.matchings_by_test_case[configuration.display_name()][test_case.name]  # type: ignore
        self.locators_by_test_case[test_case.name] = [ts.locator for ts, _, _ in inferences]  # type: ignore

        # compute nested metrics per class
        labels = {gt.label for _, gts, _ in inferences for gt in gts.bboxes} | {
            inf.label for _, _, infs in inferences for inf in infs.bboxes
        }
        per_class_metrics: List[ClassMetricsPerTestCase] = []
        for label in sorted(labels):
            metrics_per_class = self.compute_aggregate_label_metrics(all_bbox_matches, label, thresholds)
            per_class_metrics.append(metrics_per_class)

        return self.test_case_metrics(per_class_metrics, metrics)

    def compute_test_case_plots(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],  # type: ignore
        metrics: List[TestSampleMetrics],  # type: ignore
        configuration: Optional[ThresholdConfiguration] = None,  # type: ignore
    ) -> Optional[List[Plot]]:
        assert configuration is not None, "must specify configuration"
        thresholds = self.get_confidence_thresholds(configuration)
        all_bbox_matches = self.matchings_by_test_case[configuration.display_name()][test_case.name]  # type: ignore

        # clean matching for confusion matrix
        match_matched = [
            (gt, inf) for match in all_bbox_matches for gt, inf in match.matched if inf.score >= thresholds[inf.label]
        ]
        match_unmatched_gt = [
            (gt, inf)
            for match in all_bbox_matches
            for gt, inf in match.unmatched_gt
            if inf is not None and inf.score >= thresholds[inf.label]
        ]
        confusion_matrix_matchings = [
            MulticlassInferenceMatches(
                matched=match_matched,
                unmatched_gt=match_unmatched_gt,
                unmatched_inf=[],
            ),
        ]

        plots: List[Plot] = []
        plots.extend(
            filter(
                None,
                [
                    compute_f1_plot_multiclass(all_bbox_matches),
                    compute_pr_plot_multiclass(all_bbox_matches),
                    compute_confusion_matrix_plot(confusion_matrix_matchings),
                ],
            ),
        )

        return plots

    def test_suite_metrics(self, unique_locators: Set[str], average_precisions: List[float]) -> TestSuiteMetrics:
        return TestSuiteMetrics(
            n_images=len(unique_locators),
            mean_AP=np.mean(average_precisions) if average_precisions else 0.0,  # type: ignore
        )

    def compute_test_suite_metrics(
        self,
        test_suite: TestSuite,
        metrics: List[Tuple[TestCase, TestCaseMetrics]],  # type: ignore
        configuration: Optional[ThresholdConfiguration] = None,  # type: ignore
    ) -> TestSuiteMetrics:
        assert configuration is not None, "must specify configuration"
        unique_locators = {
            locator for tc, _ in metrics for locator in self.locators_by_test_case[tc.name]  # type: ignore
        }
        average_precisions = [tcm.mean_AP for _, tcm in metrics]
        return self.test_suite_metrics(unique_locators, average_precisions)

    def get_confidence_thresholds(self, configuration: ThresholdConfiguration) -> Dict[str, float]:
        if configuration.threshold_strategy == "F1-Optimal":
            return self.threshold_cache[configuration.display_name()]
        else:
            return defaultdict(lambda: configuration.threshold_strategy)  # type: ignore



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict
from typing import List
from typing import Union

from pydantic import field_validator
from pydantic import model_validator
from pydantic.dataclasses import dataclass

from kolena._api.v2._api import GeneralFieldFilter
from kolena._api.v2._testing import SimpleRange
from kolena._api.v2._testing import StratifyFieldSpec


@dataclass(frozen=True)
class CategoricalValue:
    value: Union[str, bool, int, float, None]


@dataclass(frozen=True)
class RangeValue:
    value: SimpleRange


@dataclass(frozen=True)
class QuantileNumericalValue:
    index: int


@dataclass(frozen=True)
class TestCase:
    name: str
    stratification: List[Union[QuantileNumericalValue, CategoricalValue, RangeValue]]
    # QuantileNumericalValue must come before CategoricalValue in Union;
    #   otherwise, from_dict will cast dicts like {value:[undefined], index:[some_value]}
    #   as CategoricalValue since [undefined] is treated as None


@dataclass(frozen=True)
class Stratification:
    name: str
    stratify_fields: List[StratifyFieldSpec]
    test_cases: List[TestCase]
    filters: Union[Dict[str, GeneralFieldFilter], None] = None

    @field_validator("test_cases")
    @classmethod
    def test_case_name_unique(cls, test_cases: List[TestCase]) -> List[TestCase]:
        if len(test_cases) > len({test_case.name for test_case in test_cases}):
            raise ValueError("Test case name must be unique.")
        return test_cases

    @model_validator(mode="after")
    def validate_stratify_field_or_filter(self) -> "Stratification":
        if not self.stratify_fields and not self.filters:
            raise ValueError("Must provide one of 'stratify_fields' or 'filters'")
        return self



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
from typing import Optional
from typing import Type
from urllib.parse import quote

import dacite

from kolena._api.v1.generic import Workflow as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.pydantic_v1.dataclasses import dataclass
from kolena._utils.state import kolena_initialized
from kolena.workflow import GroundTruth
from kolena.workflow import Inference
from kolena.workflow import TestSample
from kolena.workflow.ground_truth import _validate_ground_truth_type
from kolena.workflow.inference import _validate_inference_type
from kolena.workflow.test_sample import _validate_test_sample_type

_RESERVED_WORKFLOW_NAMES = {"fr", "classification", "detection", "keypoints"}


@dataclass(frozen=True)
class EvaluatorRoleConfig:
    #: evaluator role arn
    job_role_arn: str

    #: external_id to be used when assume_role
    external_id: str

    #: role_arn to be assume-into
    assume_role_arn: str


@dataclass(frozen=True)
class RemoteEvaluator:
    """
    Remote evaluator for workflows built with ``kolena.workflow``.
    """

    #: The name of the workflow
    workflow: str

    #: The name of the evaluator
    name: str

    #: The full qualified docker image name
    image: Optional[str] = None

    #: Time the remote evaluator is registered
    created: Optional[str] = None

    #: Sensitive data registered with the evaluator, only included if requested explicitly
    secret: Optional[str] = None

    #: AWS assume role configurations for the evaluator job if configured
    aws_role_config: Optional[EvaluatorRoleConfig] = None

    @classmethod
    def _from_api_response(cls, workflow: str, response: API.EvaluatorResponse) -> "RemoteEvaluator":
        aws_role_config = None
        if response.aws_role_config:
            aws_role_config = EvaluatorRoleConfig(
                job_role_arn=response.aws_role_config.job_role_arn,
                external_id=response.aws_role_config.external_id,
                assume_role_arn=response.aws_role_config.assume_role_arn,
            )
        return cls(
            workflow=workflow,
            name=response.name,
            image=response.image,
            created=response.created,
            secret=response.secret,
            aws_role_config=aws_role_config,
        )


@dataclass(frozen=True)
class Workflow:
    """The definition of a workflow and its associated types."""

    name: str
    """The name of the workflow. Should be unique, meaningful, and human-readable."""

    test_sample_type: Type[TestSample]
    """The [`TestSample`][kolena.workflow.TestSample] type for the workflow, using one of the builtin test sample types
    (e.g. [`Image`][kolena.workflow.Image]) and extending as necessary with additional fields."""

    ground_truth_type: Type[GroundTruth]
    """The custom [`GroundTruth`][kolena.workflow.GroundTruth] type for the workflow."""

    inference_type: Type[Inference]
    """The custom [`Inference`][kolena.workflow.Inference] type for the workflow."""

    def __post_init__(self) -> None:
        object.__setattr__(self, "name", self.name.strip())
        if self.name == "":
            raise ValueError("invalid zero-length name provided")
        if self.name.lower() in _RESERVED_WORKFLOW_NAMES:
            raise ValueError(f"invalid reserved name '{self.name}' provided")

        _validate_test_sample_type(self.test_sample_type)
        _validate_ground_truth_type(self.test_sample_type, self.ground_truth_type)
        _validate_inference_type(self.test_sample_type, self.inference_type)

    @kolena_initialized
    def register_evaluator(
        self,
        evaluator_name: str,
        image: str,
        secret: Optional[str] = None,
        aws_assume_role: Optional[str] = None,
    ) -> RemoteEvaluator:
        """
        Convenience method to register evaluator for the workflow.

        See [`register_evaluator`][kolena.workflow.workflow.register_evaluator] for details.
        """
        return register_evaluator(
            workflow=self.name,
            evaluator_name=evaluator_name,
            image=image,
            secret=secret,
            aws_assume_role=aws_assume_role,
        )

    @kolena_initialized
    def get_evaluator(self, evaluator_name: str, include_secret: bool = False) -> RemoteEvaluator:
        """
        Get the docker image registered for the evaluator.

        See [`get_evaluator`][kolena.workflow.workflow.get_evaluator] for details.
        """

        return get_evaluator(self.name, evaluator_name, include_secret)


@kolena_initialized
def register_evaluator(
    workflow: str,
    evaluator_name: str,
    image: str,
    secret: Optional[str] = None,
    aws_assume_role: Optional[str] = None,
) -> RemoteEvaluator:
    """
    Register a docker image for the evaluator of the workflow. This enables metrics evaluation to be run in
    kolena platform with the given image.

    If the evaluator needs to use sensitive data during its computation, pass it in ``secret``. The content would be
    store securely in kolena platform. At runtime, the evaluator can access the content in environment variable
    `KOLENA_EVALUATOR_SECRET`.

    If the evaluator needs to interact with AWS APIs, specify the role it would use in ``aws_assume_role``. The
    runtime would be executed by a Kolena-side AWS role that is configured to assume the ``aws_assume_role`` using a
    randomly generated `external_id`. The AWS role arn from Kolena and the `external_id` would be returned in the
    response `aws_assume_role` field. The only configuration on user side is to add Kolena AWS role to
    ``aws_assume_role``'s trust policy. At runtime, the evaluator code can get the external_id in environment
    variable `KOLENA_EVALUATOR_EXTERNAL_ID` and the ``aws_assume_role`` would be passed in
    `KOLENA_EVALUATOR_ASSUME_ROLE_ARN`.

    An example to assume role at evaluator runtime is shown below::

        external_id = os.environ["KOLENA_EVALUATOR_EXTERNAL_ID"]
        assume_role_arn = os.environ["KOLENA_EVALUATOR_ASSUME_ROLE_ARN"]  # this is the role to assume

        sts = boto3.client("sts")
        response = sts.assume_role(
            RoleArn=assume_role_arn,
            RoleSessionName="metrics-evaluator",
            ExternalId=external_id,
        )
        credentials = response["Credentials"]
        client = boto3.client(
            "s3",
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
        )
        buckets = client.list_buckets()

    :param evaluator_name: name of the evaluator; must match evaluator name generated by image
    :param image: fully qualified docker image name, i.e. <repository-url>/<image>:<tag>
    :param secret: sensitive data in string format
    :param aws_assume_role: AWS role that the evaluator would use to access AWS APIs
    :return: registered evaluator data
    """

    response = krequests.post(
        API.Path.EVALUATOR.value,
        json=dict(workflow=workflow, image=image, name=evaluator_name, secret=secret, aws_assume_role=aws_assume_role),
    )
    krequests.raise_for_status(response)

    log.info(f"Image {image} successfully registered for evaluator {evaluator_name}.")

    result = dacite.from_dict(API.EvaluatorResponse, response.json())
    return RemoteEvaluator._from_api_response(workflow, result)


@kolena_initialized
def list_evaluators(workflow: str) -> List[RemoteEvaluator]:
    """
    List all evaluators registered for a given workflow.

    :param workflow: workflow name
    :return: list of registered evaluators
    """

    response = krequests.get(f"{API.Path.EVALUATOR.value}/{quote(workflow)}")
    krequests.raise_for_status(response)

    return [
        RemoteEvaluator(workflow=workflow, name=evaluator.name, image=evaluator.image, created=evaluator.created)
        for evaluator in dacite.from_dict(API.ListEvaluatorsResponse, response.json()).evaluators
    ]


@kolena_initialized
def get_evaluator(workflow: str, evaluator_name: str, include_secret: bool = False) -> RemoteEvaluator:
    """
    Get the latest evaluator registered for the ``workflow`` and ``evaluator_name``.

    :param workflow: workflow name
    :param evaluator_name: evaluator name
    :param include_secret: retrieve secret registered with the evaluator or not
    :return: remote evaluator
    """

    response = krequests.get(
        endpoint_path=f"{API.Path.EVALUATOR.value}/{quote(workflow)}/{quote(evaluator_name)}",
        params=dict(include_secret=include_secret),
    )
    krequests.raise_for_status(response)

    result = dacite.from_dict(API.EvaluatorResponse, response.json())
    return RemoteEvaluator._from_api_response(workflow, result)



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from pydantic import conlist
from pydantic import constr
from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class Scopes:
    # None means no scope, empty array means selecting nothing
    datapoint: Optional[conlist(constr(min_length=1))] = None
    result: Optional[conlist(constr(min_length=1))] = None
    human_evaluation: Optional[conlist(constr(min_length=1))] = None

    def __hash__(self) -> int:
        return hash(
            (
                tuple(self.datapoint) if self.datapoint is not None else None,
                tuple(self.result) if self.result is not None else None,
                tuple(self.human_evaluation) if self.human_evaluation is not None else None,
            ),
        )



# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict
from typing import Literal

from pydantic import constr
from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class ResultContext:
    model_id: int
    eval_config_id: int


Contexts = Dict[str, ResultContext]


@dataclass(frozen=True)
class Dsl:
    contexts: Contexts
    query: constr(min_length=1, strip_whitespace=True)

    def __hash__(self) -> int:
        return hash(
            (
                tuple(self.contexts),
                self.query,
            ),
        )


Mode = Literal["sort", "filter", "derived_field"]


@dataclass(frozen=True)
class ValidateDslRequest:
    dsl: Dsl
    mode: Mode
